# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from glob import glob
from inspect import isclass
from typing import Any, Dict, Optional, Tuple, Type, TypeVar, Union

from datumaro.components.annotation import Bbox, RleMask
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import (
    DatasetImportError,
    InvalidAnnotationError,
    InvalidFieldTypeError,
    MissingFieldError,
)
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image
from datumaro.util import NOTSET, parse_json_file

T = TypeVar("T")


def parse_field(
    ann: Dict[str, Any],
    key: str,
    cls: Union[Type[T], Tuple[Type, ...]],
    default: Any = NOTSET,
) -> Any:
    value = ann.get(key, NOTSET)
    if value is NOTSET:
        if default is not NOTSET:
            return default
        raise MissingFieldError(key)
    elif not isinstance(value, cls):
        cls = (cls,) if isclass(cls) else cls
        raise InvalidFieldTypeError(
            key, actual=str(type(value)), expected=tuple(str(t) for t in cls)
        )
    return value


class SegmentAnythingBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isdir(path):
            raise DatasetImportError(f"path {path} must be directory.")
        self._path = path

        super().__init__(subset=subset, ctx=ctx)
        self._items = self._load_items()

    def _load_items(self):
        pbar = self._ctx.progress_reporter
        items = []
        for annotation_file in pbar.iter(
            glob(osp.join(self._path, "*.json")),
            desc=f"Parsing data in {osp.basename(self._path)}",
        ):
            image_id = None
            annotations = []
            item_kwargs = {
                "id": None,
                "subset": self._subset,
                "media": None,
                "annotations": [],
                "attributes": {},
            }

            try:
                contents = parse_json_file(annotation_file)
                image_info = contents["image"]
                annotations = contents["annotations"]

                image_id = parse_field(image_info, "image_id", int)
                item_kwargs["attributes"]["id"] = image_id

                image_size = (
                    parse_field(image_info, "height", int, default=None),
                    parse_field(image_info, "width", int, default=None),
                )
                if any(i is None for i in image_size):
                    image_size = None
                file_name = parse_field(image_info, "file_name", str)

                item_kwargs["id"] = osp.splitext(file_name)[0]
                item_kwargs["media"] = Image.from_file(
                    path=osp.join(self._path, file_name), size=image_size
                )
            except Exception as e:
                self._ctx.error_policy.report_item_error(e, item_id=(image_id, self._subset))

            try:
                for annotation in annotations:
                    anno_id = parse_field(annotation, "id", int)
                    attributes = {
                        "predicted_iou": parse_field(
                            annotation,
                            "predicted_iou",
                            float,
                            0.0,
                        ),
                        "stability_score": parse_field(
                            annotation,
                            "stability_score",
                            float,
                            0.0,
                        ),
                        "point_coords": parse_field(
                            annotation,
                            "point_coords",
                            list,
                            [[]],
                        ),
                        "crop_box": parse_field(annotation, "crop_box", list, []),
                    }

                    group = anno_id  # make sure all tasks' annotations are merged

                    segmentation = parse_field(annotation, "segmentation", dict, None)
                    if segmentation is None:
                        raise InvalidAnnotationError("'segmentation' label is not found.")
                    item_kwargs["annotations"].append(
                        RleMask(
                            rle=segmentation,
                            label=None,
                            id=anno_id,
                            attributes=attributes,
                            group=group,
                        )
                    )

                    bbox = parse_field(annotation, "bbox", list, None)
                    if bbox is None:
                        bbox = item_kwargs["annotations"][-1].get_bbox().tolist()

                    if len(bbox) > 0:
                        if len(bbox) != 4:
                            raise InvalidAnnotationError(
                                f"Bbox has wrong value count {len(bbox)}. Expected 4 values."
                            )
                        x, y, w, h = bbox
                        item_kwargs["annotations"].append(
                            Bbox(
                                x,
                                y,
                                w,
                                h,
                                label=None,
                                id=anno_id,
                                attributes=attributes,
                                group=group,
                            )
                        )
            except Exception as e:
                self._ctx.error_policy.report_annotation_error(e, item_id=(image_id, self._subset))

            try:
                items.append(DatasetItem(**item_kwargs))
                for ann in item_kwargs["annotations"]:
                    self._ann_types.add(ann.type)
            except Exception as e:
                self._ctx.error_policy.report_item_error(e, item_id=(image_id, self._subset))

        return items



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
from typing import Dict, List, Optional

import pyarrow as pa

from datumaro.components.errors import DatasetImportError
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer

from .format import DatumaroArrow

__all__ = ["ArrowImporter"]


class ArrowImporter(Importer):
    _FORMAT_EXT = ".arrow"

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> Optional[FormatDetectionConfidence]:
        if context.root_path.endswith(".arrow"):
            cls._verify_datumaro_arrow_format(context.root_path)
        else:
            for arrow_file in context.require_files("*.arrow"):
                with context.probe_text_file(
                    arrow_file,
                    f"{arrow_file} is not Datumaro arrow format.",
                    is_binary_file=True,
                ) as f:
                    f.close()
                    cls._verify_datumaro_arrow_format(os.path.join(context.root_path, arrow_file))

    @classmethod
    def find_sources(cls, path: str) -> List[Dict]:
        def _filter(path: str) -> bool:
            try:
                cls._verify_datumaro_arrow_format(path)
                return True
            except DatasetImportError:
                return False

        return cls._find_sources_recursive(
            path=path,
            ext=cls._FORMAT_EXT,
            extractor_name=cls.NAME,
            file_filter=_filter,
            max_depth=0,
        )

    @classmethod
    def find_sources_with_params(cls, path: str, **extra_params) -> List[Dict]:
        sources = cls.find_sources(path)
        # Merge sources into one config but multiple file_paths
        return [
            {
                "url": path,
                "format": cls.NAME,
                "options": {"file_paths": [source["url"] for source in sources]},
            }
        ]

    @staticmethod
    def _verify_datumaro_arrow_format(file: str) -> None:
        with pa.memory_map(file, "r") as mm_file:
            with pa.ipc.open_file(mm_file) as reader:
                schema = reader.schema
        DatumaroArrow.check_signature(schema.metadata.get(b"signature", b"").decode())
        DatumaroArrow.check_version(schema.metadata.get(b"version", b"").decode())
        DatumaroArrow.check_schema(schema)

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._FORMAT_EXT]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Dict, Optional, Sequence, Type

from datumaro.components.annotation import Annotation
from datumaro.components.dataset_base import IDataset
from datumaro.components.dataset_item_storage import (
    DatasetItemStorage,
    DatasetItemStorageDatasetView,
)
from datumaro.components.media import MediaElement

__all__ = ["IMatcherContext", "IMergerContext"]


class IMatcherContext(ABC):
    @abstractmethod
    def get_any_label_name(self, ann: Annotation, label_id: int) -> str:
        raise NotImplementedError


class IMergerContext(IMatcherContext):
    @abstractmethod
    def merge_infos(self, sources: Sequence[IDataset]) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def merge_categories(self, sources: Sequence[IDataset]) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def merge_media_types(self, sources: Sequence[IDataset]) -> Optional[Type[MediaElement]]:
        raise NotImplementedError

    @abstractmethod
    def merge(self, sources: Sequence[IDataset]) -> DatasetItemStorage:
        raise NotImplementedError

    @abstractmethod
    def __call__(self, *datasets: IDataset) -> DatasetItemStorageDatasetView:
        raise NotImplementedError



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from datumaro.util.tf_util import import_tf

import_tf()  # prevent TF loading and potential interpreter crash

import openvino.tools.accuracy_checker.representation as ac

import datumaro.components.dataset_base as dm
from datumaro.util.annotation_util import softmax


def import_predictions(predictions):
    # Convert Accuracy checker predictions to Datumaro annotations

    anns = []

    for pred in predictions:
        anns.extend(import_prediction(pred))

    return anns


def import_prediction(pred):
    if isinstance(pred, ac.ClassificationPrediction):
        scores = softmax(pred.scores)
        return (
            dm.Label(label_id, attributes={"score": float(score)})
            for label_id, score in enumerate(scores)
        )
    elif isinstance(pred, ac.ArgMaxClassificationPrediction):
        return (dm.Label(int(pred.label)),)
    elif isinstance(pred, ac.CharacterRecognitionPrediction):
        return (dm.Label(int(pred.label)),)
    elif isinstance(pred, (ac.DetectionPrediction, ac.ActionDetectionPrediction)):
        return (
            dm.Bbox(x0, y0, x1 - x0, y1 - y0, int(label_id), attributes={"score": float(score)})
            for label_id, score, x0, y0, x1, y1 in zip(
                pred.labels, pred.scores, pred.x_mins, pred.y_mins, pred.x_maxs, pred.y_maxs
            )
        )
    elif isinstance(pred, ac.DepthEstimationPrediction):
        return (dm.Mask(pred.depth_map),)  # 2d floating point mask
    # elif isinstance(pred, ac.HitRatioPrediction):
    #     -
    elif isinstance(pred, ac.ImageInpaintingPrediction):
        return (dm.Mask(pred.value),)  # an image
    # elif isinstance(pred, ac.MultiLabelRecognitionPrediction):
    #     -
    # elif isinstance(pred, ac.MachineTranslationPrediction):
    #     -
    # elif isinstance(pred, ac.QuestionAnsweringPrediction):
    #     -
    # elif isinstance(pred, ac.PoseEstimation3dPrediction):
    #     -
    # elif isinstance(pred, ac.PoseEstimationPrediction):
    #     -
    # elif isinstance(pred, ac.RegressionPrediction):
    #     -
    else:
        raise NotImplementedError("Can't convert %s" % type(pred))



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import re
from enum import IntEnum
from typing import Dict


class YoloPath:
    DEFAULT_SUBSET_NAME = "train"
    SUBSET_NAMES = ["train", "valid"]
    RESERVED_CONFIG_KEYS = {"backup", "classes", "names"}

    @staticmethod
    def _parse_config(path: str) -> Dict[str, str]:
        with open(path, "r", encoding="utf-8") as f:
            config_lines = f.readlines()

        config = {}

        for line in config_lines:
            match = re.match(r"^\s*(\w+)\s*=\s*(.+)$", line)
            if not match:
                continue

            key = match.group(1)
            value = match.group(2)
            config[key] = value

        return config


class YoloLoosePath:
    NAMES_FILE = "obj.names"


class YoloUltralyticsPath:
    META_FILE = "data.yaml"


class YoloFormatType(IntEnum):
    yolo_strict = 0
    yolo_loose = 1
    yolo_ultralytics = 2



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Tuple

import cv2

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image
from datumaro.plugins.openvino_plugin.samples.utils import (
    create_bboxes_with_rescaling,
    rescale_img_keeping_aspect_ratio,
)

__all__ = ["OTXYoloXModelInterpreter"]


class OTXYoloXModelInterpreter(IModelInterpreter):
    h_model = 416
    w_model = 416

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data
        output = rescale_img_keeping_aspect_ratio(img, self.h_model, self.w_model)

        # From BGR to RGB
        output.image = cv2.cvtColor(output.image, cv2.COLOR_BGR2RGB)
        # From HWC to CHW
        output.image = output.image.transpose(2, 0, 1)

        return output.image, output.scale

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        scale = info
        r_scale = 1 / scale
        return create_bboxes_with_rescaling(pred["boxes"], pred["labels"], r_scale)

    def get_categories(self):
        return None



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import os.path as osp

from datumaro.util.scope import scope_add, scoped

from ....util import MultilineFormatter
from ....util.project import load_project


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Get project info",
        description="""
        Outputs project info - information about plugins,
        sources, build tree, models and revisions.|n
        |n
        Examples:|n
        - Print project info for the current working tree:|n |n
        |s|s%(prog)s|n
        |n
        - Print project info for the previous revision:|n |n
        |s|s%(prog)s HEAD~1
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "revision", default="", nargs="?", help="Target revision (default: current working tree)"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=info_command)

    return parser


def get_sensitive_args():
    return {
        info_command: ["project_dir", "revision"],
    }


@scoped
def info_command(args):
    project = scope_add(load_project(args.project_dir))
    rev = project.get_rev(args.revision)
    env = rev.env

    print("Project:")
    print("  location:", project._root_dir)
    print("Plugins:")
    print("  extractors:", ", ".join(sorted(set(env.extractors) | set(env.importers))))
    print("  exporters:", ", ".join(env.exporters))
    print("  launchers:", ", ".join(env.launchers))

    print("Models:")
    for model_name, model in project.models.items():
        print("  model '%s':" % model_name)
        print("    type:", model.launcher)

    print("Sources:")
    for source_name, source in rev.sources.items():
        print("  '%s':" % source_name)
        print("    format:", source.format)
        print("    url:", osp.abspath(source.url) if source.url else "")
        print(
            "    location:",
            osp.abspath(osp.join(project.source_data_dir(source_name), source.path)),
        )
        print("    options:", source.options)

        print("    stages:")
        for stage in rev.build_targets[source_name].stages:
            print("      '%s':" % stage.name)
            print("        type:", stage.type)
            print("        hash:", stage.hash)
            print("        cached:", project.is_obj_cached(stage.hash) if stage.hash else "n/a")
            if stage.kind:
                print("        kind:", stage.kind)
            if stage.params:
                print("        parameters:", stage.params)

    return 0



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT
from multiprocessing.pool import ThreadPool
from typing import Iterator, List, Optional

import numpy as np

from datumaro.components.annotation import Annotation, AnnotationType, LabelCategories
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetBase, DatasetItem, IDataset
from datumaro.components.launcher import Launcher
from datumaro.util import is_method_redefined, take_by
from datumaro.util.multi_procs_util import consumer_generator


class Transform(DatasetBase, CliPlugin):
    """
    A base class for dataset transformations that change dataset items
    or their annotations.
    """

    @staticmethod
    def wrap_item(item, **kwargs):
        return item.wrap(**kwargs)

    def __init__(self, extractor: IDataset):
        super().__init__()

        self._extractor = extractor

    def categories(self):
        return self._extractor.categories()

    def subsets(self):
        if self._subsets is None:
            self._subsets = set(self._extractor.subsets())
        return super().subsets()

    def __len__(self):
        assert self._length in {None, "parent"} or isinstance(self._length, int)
        if (
            self._length is None
            and not is_method_redefined("__iter__", Transform, self)
            or self._length == "parent"
        ):
            self._length = len(self._extractor)
        return super().__len__()

    def media_type(self):
        return self._extractor.media_type()

    def infos(self):
        return self._extractor.infos()


class ItemTransform(Transform):
    def transform_item(self, item: DatasetItem) -> Optional[DatasetItem]:
        """
        Returns a modified copy of the input item.

        Avoid changing and returning the input item, because it can lead to
        unexpected problems. Use wrap_item() or item.wrap() to simplify copying.
        """

        raise NotImplementedError()

    def __iter__(self):
        for item in self._extractor:
            item = self.transform_item(item)
            if item is not None:
                yield item


class ModelTransform(Transform):
    """A transformation class for applying a model's inference to dataset items.

    This class takes an dataset, a launcher, and other optional parameters
    to transform the dataset item from the model outputs by the launcher.
    It can process items using multiple processes if specified, making it suitable for
    parallelized inference tasks.

    Parameters:
        extractor: The dataset extractor to obtain items from.
        launcher: The launcher responsible for model inference.
        batch_size: The batch size for processing items. Default is 1.
        append_annotation: Whether to append inference annotations to existing annotations.
            Default is False.
        num_workers: The number of worker threads to use for parallel inference.
            Set to 0 for single-process mode. Default is 0.
    """

    def __init__(
        self,
        extractor: IDataset,
        launcher: Launcher,
        batch_size: int = 1,
        append_annotation: bool = False,
        num_workers: int = 0,
    ):
        super().__init__(extractor)
        self._launcher = launcher
        self._batch_size = batch_size
        self._append_annotation = append_annotation
        if not (isinstance(num_workers, int) and num_workers >= 0):
            raise ValueError(
                f"num_workers should be a non negative integer, but it is {num_workers}"
            )
        self._num_workers = num_workers

    def __iter__(self) -> Iterator[DatasetItem]:
        if self._num_workers == 0:
            return self._iter_single_proc()
        return self._iter_multi_procs()

    def _iter_multi_procs(self):
        with ThreadPool(processes=self._num_workers) as pool:

            def _producer_gen():
                for batch in take_by(self._extractor, self._batch_size):
                    future = pool.apply_async(
                        func=self._process_batch,
                        args=(batch,),
                    )
                    yield future

            with consumer_generator(producer_generator=_producer_gen()) as consumer_gen:
                for future in consumer_gen:
                    for item in future.get():
                        yield item

    def _iter_single_proc(self) -> Iterator[DatasetItem]:
        for batch in take_by(self._extractor, self._batch_size):
            for item in self._process_batch(batch=batch):
                yield item

    def _process_batch(
        self,
        batch: List[DatasetItem],
    ) -> List[DatasetItem]:
        inference = self._launcher.launch(
            batch=[item for item in batch if self._launcher.type_check(item)]
        )

        for annotations in inference:
            self._check_annotations(annotations)

        return [
            self.wrap_item(
                item,
                annotations=item.annotations + annotations
                if self._append_annotation
                else annotations,
            )
            for item, annotations in zip(batch, inference)
        ]

    def get_subset(self, name):
        subset = self._extractor.get_subset(name)
        return __class__(subset, self._launcher, self._batch_size)

    def infos(self):
        launcher_override = self._launcher.infos()
        if launcher_override is not None:
            return launcher_override
        return self._extractor.infos()

    def categories(self):
        launcher_override = self._launcher.categories()
        if launcher_override is not None:
            return launcher_override
        return self._extractor.categories()

    def transform_item(self, item):
        inputs = np.expand_dims(item.media, axis=0)
        annotations = self._launcher.launch(inputs)[0]
        return self.wrap_item(item, annotations=annotations)

    def _check_annotations(self, annotations: List[Annotation]):
        labels_count = len(self.categories().get(AnnotationType.label, LabelCategories()).items)

        for ann in annotations:
            label = getattr(ann, "label", None)
            if label is None:
                continue

            if label not in range(labels_count):
                raise Exception(
                    "Annotation has unexpected label id %s, "
                    "while there is only %s defined labels." % (label, labels_count)
                )



# Copyright (C) 2020-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import inspect
import logging as log
import os
import os.path as osp
import warnings
from contextlib import contextmanager
from copy import copy, deepcopy
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    overload,
)

from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    LabelCategories,
    TabularCategories,
)
from datumaro.components.config_model import Source
from datumaro.components.dataset_base import (
    DEFAULT_SUBSET_NAME,
    CategoriesInfo,
    DatasetBase,
    DatasetInfo,
    DatasetItem,
    IDataset,
)
from datumaro.components.dataset_item_storage import DatasetItemStorageDatasetView
from datumaro.components.dataset_storage import DatasetPatch, DatasetStorage, StreamDatasetStorage
from datumaro.components.environment import DEFAULT_ENVIRONMENT, Environment
from datumaro.components.errors import (
    DatasetImportError,
    DatumaroError,
    MultipleFormatsMatchError,
    NoMatchingFormatsError,
    StreamedItemError,
    UnknownFormatError,
)
from datumaro.components.exporter import ExportContext, Exporter, ExportErrorPolicy, _ExportFail
from datumaro.components.filter import (
    UserFunctionAnnotationsFilter,
    UserFunctionDatasetFilter,
    XPathAnnotationsFilter,
    XPathDatasetFilter,
)
from datumaro.components.importer import ImportContext, ImportErrorPolicy, _ImportFail
from datumaro.components.launcher import Launcher
from datumaro.components.media import Image, MediaElement
from datumaro.components.merge import DEFAULT_MERGE_POLICY
from datumaro.components.progress_reporting import NullProgressReporter, ProgressReporter
from datumaro.components.transformer import ItemTransform, ModelTransform, Transform
from datumaro.util.log_utils import logging_disabled
from datumaro.util.meta_file_util import load_hash_key
from datumaro.util.os_util import rmtree
from datumaro.util.scope import on_error_do, scoped

DEFAULT_FORMAT = "datumaro"

__all__ = ["Dataset", "eager_mode"]


class DatasetSubset(IDataset):  # non-owning view
    def __init__(self, parent: Dataset, name: str):
        super().__init__()
        self.parent = parent
        self.name = name

    def __iter__(self):
        yield from self.parent._data.get_subset(self.name)

    def __len__(self):
        subset: DatasetItemStorageDatasetView.Subset = self.parent._data.get_subset(self.name)

        return len(subset)

    def put(self, item):
        return self.parent.put(item, subset=self.name)

    def get(self, id, subset=None):
        assert (subset or DEFAULT_SUBSET_NAME) == (self.name or DEFAULT_SUBSET_NAME)
        return self.parent.get(id, subset=self.name)

    def remove(self, id, subset=None):
        assert (subset or DEFAULT_SUBSET_NAME) == (self.name or DEFAULT_SUBSET_NAME)
        return self.parent.remove(id, subset=self.name)

    def get_subset(self, name):
        assert (name or DEFAULT_SUBSET_NAME) == (self.name or DEFAULT_SUBSET_NAME)
        return self

    def subsets(self):
        if (self.name or DEFAULT_SUBSET_NAME) == DEFAULT_SUBSET_NAME:
            return self.parent.subsets()
        return {self.name: self}

    def infos(self):
        return self.parent.infos()

    def categories(self):
        return self.parent.categories()

    def media_type(self):
        return self.parent.media_type()

    def ann_types(self):
        return self.parent.ann_types()

    def get_annotated_items(self):
        return sum(bool(s.annotations) for s in self.parent._data.get_subset(self.name))

    def get_annotations(self):
        annotations_by_type = {t.name: {"count": 0} for t in AnnotationType}
        for item in self.parent._data.get_subset(self.name):
            for ann in item.annotations:
                annotations_by_type[ann.type.name]["count"] += 1
        return sum(t["count"] for t in annotations_by_type.values())

    def as_dataset(self) -> Dataset:
        dataset = Dataset.from_extractors(self, env=self.parent.env)
        dataset._format = self.parent._format
        dataset._source_path = self.parent._source_path
        return dataset


class Dataset(IDataset):
    """
    Represents a dataset, contains metainfo about labels and dataset items.
    Provides iteration and access options to dataset elements.

    By default, all operations are done lazily, it can be changed by
    modifying the `eager` property and by using the `eager_mode`
    context manager.

    Dataset is supposed to have a single media type for its items. If the
    dataset is filled manually or from extractors, and media type does not
    match, an error is raised.
    """

    _global_eager: bool = False
    _stream = False

    @classmethod
    def from_iterable(
        cls,
        iterable: Iterable[DatasetItem],
        infos: Optional[DatasetInfo] = None,
        categories: Union[CategoriesInfo, List[str], None] = None,
        *,
        env: Optional[Environment] = None,
        media_type: Type[MediaElement] = Image,
        ann_types: Optional[Set[AnnotationType]] = [],
    ) -> Dataset:
        """
        Creates a new dataset from an iterable object producing dataset items -
        a generator, a list etc. It is a convenient way to create and fill
        a custom dataset.

        Parameters:
            iterable: An iterable which returns dataset items
            infos: A dictionary of the dataset specific information
            categories: A simple list of labels or complete information
                about labels. If not specified, an empty list of labels
                is assumed.
            media_type: Media type for the dataset items. If the sequence
                contains items with mismatching media type, an error is
                raised during caching
            env: A context for plugins, which will be used for this dataset.
                If not specified, the builtin plugins will be used.

        Returns:
            dataset: A new dataset with specified contents
        """

        if infos is None:
            infos = {}

        # TODO: remove the default value for media_type
        # https://github.com/openvinotoolkit/datumaro/issues/675

        if isinstance(categories, list):
            categories = {AnnotationType.label: LabelCategories.from_iterable(categories)}

        if not categories:
            categories = {}

        class _extractor(DatasetBase):
            def __init__(self):
                super().__init__(
                    length=len(iterable) if hasattr(iterable, "__len__") else None,
                    media_type=media_type,
                    ann_types=ann_types,
                )

            def __iter__(self):
                return iter(iterable)

            def infos(self):
                return infos

            def categories(self):
                return categories

        return cls.from_extractors(_extractor(), env=env)

    @classmethod
    def from_extractors(
        cls,
        *sources: IDataset,
        env: Optional[Environment] = None,
        merge_policy: str = DEFAULT_MERGE_POLICY,
    ) -> Dataset:
        """
        Creates a new dataset from one or several `Extractor`s.

        In case of a single input, creates a lazy wrapper around the input.
        In case of several inputs, merges them and caches the resulting
        dataset.

        Parameters:
            sources: one or many input extractors
            env: A context for plugins, which will be used for this dataset.
                If not specified, the builtin plugins will be used.
            merge_policy: Policy on how to merge multiple datasets.
                Possible options are "exact", "intersect", and "union".

        Returns:
            dataset: A new dataset with contents produced by input extractors
        """

        if len(sources) == 1:
            source = sources[0]
            dataset = cls(source=source, env=env)
        else:
            from datumaro.components.hl_ops import HLOps

            return HLOps.merge(*sources, merge_policy=merge_policy)

        return dataset

    def __init__(
        self,
        source: Optional[IDataset] = None,
        *,
        infos: Optional[DatasetInfo] = None,
        categories: Optional[CategoriesInfo] = None,
        media_type: Optional[Type[MediaElement]] = None,
        ann_types: Optional[Set[AnnotationType]] = None,
        env: Optional[Environment] = None,
    ) -> None:
        super().__init__()

        assert env is None or isinstance(env, Environment), env
        self._env = env

        self.eager = None
        self._data = DatasetStorage(
            source=source,
            infos=infos,
            categories=categories,
            media_type=media_type,
            ann_types=ann_types,
        )
        if self.is_eager:
            self.init_cache()

        self._format = DEFAULT_FORMAT
        self._source_path = None
        self._options = {}

    def __repr__(self) -> str:
        separator = "\t"
        return (
            f"Dataset\n"
            f"\tsize={len(self._data)}\n"
            f"\tsource_path={self._source_path}\n"
            f"\tmedia_type={self.media_type()}\n"
            f"\tann_types={self.ann_types()}\n"
            f"\tannotated_items_count={self.get_annotated_items()}\n"
            f"\tannotations_count={self.get_annotations()}\n"
            f"subsets\n"
            f"\t{separator.join(self.get_subset_info())}"
            f"infos\n"
            f"\t{separator.join(self.get_infos())}"
            f"categories\n"
            f"\t{separator.join(self.get_categories_info())}"
        )

    def define_infos(self, infos: DatasetInfo) -> None:
        self._data.define_infos(infos)

    def define_categories(self, categories: CategoriesInfo) -> None:
        self._data.define_categories(categories)

    def init_cache(self) -> None:
        self._data.init_cache()

    def __iter__(self) -> Iterator[DatasetItem]:
        yield from self._data

    def __len__(self) -> int:
        return len(self._data)

    def get_subset(self, name) -> DatasetSubset:
        return DatasetSubset(self, name)

    def subsets(self) -> Dict[str, DatasetSubset]:
        return {k: self.get_subset(k) for k in self._data.subsets()}

    def infos(self) -> DatasetInfo:
        return self._data.infos()

    def categories(self) -> CategoriesInfo:
        return self._data.categories()

    def media_type(self) -> Type[MediaElement]:
        return self._data.media_type()

    def ann_types(self) -> Set[AnnotationType]:
        return self._data.ann_types()

    def get(self, id: str, subset: Optional[str] = None) -> Optional[DatasetItem]:
        return self._data.get(id, subset)

    def get_annotated_items(self):
        return self._data.get_annotated_items()

    def get_annotations(self):
        return self._data.get_annotations()

    def get_datasetitem_by_path(self, path):
        if self._source_path not in path:
            path = osp.join(self._source_path, path)
        return self._data.get_datasetitem_by_path(path)

    def get_label_cat_names(self):
        return [
            label.name
            for label in self._data.categories().get(AnnotationType.label, LabelCategories())
        ]

    def get_subset_info(self) -> str:
        return (
            f"{subset_name}: # of items={len(self.get_subset(subset_name))}, "
            f"# of annotated items={self.get_subset(subset_name).get_annotated_items()}, "
            f"# of annotations={self.get_subset(subset_name).get_annotations()}\n"
            for subset_name in sorted(self.subsets().keys())
        )

    def get_infos(self) -> Tuple[str]:
        if self.infos() is not None:
            return (f"{k}: {v}\n" for k, v in self.infos().items())
        else:
            return ("\n",)

    def get_categories_info(self) -> Tuple[str]:
        category_dict = {}
        for annotation_type, category in self.categories().items():
            if isinstance(category, LabelCategories):
                category_names = list(category._indices.keys())
                category_dict[annotation_type] = category_names
            if isinstance(category, TabularCategories):
                category_names = list(category._indices_by_name.keys())
                category_dict[annotation_type] = category_names
        return (
            f"{str(annotation_type).split('.')[-1]}: {list(category_dict.get(annotation_type, []))}\n"
            for annotation_type in self.categories().keys()
        )

    def __contains__(self, x: Union[DatasetItem, str, Tuple[str, str]]) -> bool:
        if isinstance(x, DatasetItem):
            x = (x.id, x.subset)
        elif not isinstance(x, (tuple, list)):
            x = [x]
        return self.get(*x) is not None

    def put(
        self, item: DatasetItem, id: Optional[str] = None, subset: Optional[str] = None
    ) -> None:
        overrides = {}
        if id is not None:
            overrides["id"] = id
        if subset is not None:
            overrides["subset"] = subset
        if overrides:
            item = item.wrap(**overrides)

        self._data.put(item)

    def remove(self, id: str, subset: Optional[str] = None) -> None:
        self._data.remove(id, subset)

    @overload
    def filter(
        self,
        expr: str,
        *,
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ) -> Dataset:
        """
        Filters out some dataset items or annotations, using a custom filter
        expression.

        Results are stored in-place. Modifications are applied lazily.

        Args:
            expr: XPath-formatted filter expression
                (e.g. `/item[subset = 'train']`,
                `/item/annotation[label = 'cat']`)
            filter_annotations: Indicates if the filter should be
                applied to items or annotations
            remove_empty: When filtering annotations, allows to
                exclude empty items from the resulting dataset

        Returns: self
        """
        ...

    @overload
    def filter(
        self,
        filter_func: Union[
            Callable[[DatasetItem], bool], Callable[[DatasetItem, Annotation], bool]
        ],
        *,
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ) -> Dataset:
        """
        Filters out some dataset items or annotations, using a user-provided filter
        Python function.

        Results are stored in-place. Modifications are applied lazily.

        Args:
            filter_func: User-provided Python function for filtering
            filter_annotations: Indicates if the filter should be
                applied to items or annotations
            remove_empty: When filtering annotations, allows to
                exclude empty items from the resulting dataset

        Returns: self

        Example:
            - (`filter_annotations=False`) This is an example of filtering
                dataset items with images larger than 1024 pixels::

                from datumaro.components.media import Image

                def filter_func(item: DatasetItem) -> bool:
                    h, w = item.media_as(Image).size
                    return h > 1024 or w > 1024

                filtered = dataset.filter(filter_func=filter_func, filter_annotations=False)
                # No items with an image height or width greater than 1024
                filtered_items = [item for item in filtered]

            - (`filter_annotations=True`) This is an example of removing bounding boxes
                sized greater than 50% of the image size::

                from datumaro.components.media import Image
                from datumaro.components.annotation import Annotation, Bbox

                def filter_func(item: DatasetItem, ann: Annotation) -> bool:
                    # If the annotation is not a Bbox, do not filter
                    if not isinstance(ann, Bbox):
                        return False

                    h, w = item.media_as(Image).size
                    image_size = h * w
                    bbox_size = ann.h * ann.w

                    # Accept Bboxes smaller than 50% of the image size
                    return bbox_size < 0.5 * image_size

                filtered = dataset.filter(filter_func=filter_func, filter_annotations=True)
                # No bounding boxes with a size greater than 50% of their image
                filtered_items = [item for item in filtered]
        """
        ...

    def filter(
        self,
        expr_or_filter_func: Union[
            str, Callable[[DatasetItem], bool], Callable[[DatasetItem, Annotation], bool]
        ],
        *,
        filter_annotations: bool = False,
        remove_empty: bool = False,
    ) -> Dataset:
        if isinstance(expr_or_filter_func, str):
            expr = expr_or_filter_func
            return (
                self.transform(XPathAnnotationsFilter, xpath=expr, remove_empty=remove_empty)
                if filter_annotations
                else self.transform(XPathDatasetFilter, xpath=expr)
            )
        elif callable(expr_or_filter_func):
            filter_func = expr_or_filter_func
            return (
                self.transform(
                    UserFunctionAnnotationsFilter,
                    filter_func=filter_func,
                    remove_empty=remove_empty,
                )
                if filter_annotations
                else self.transform(UserFunctionDatasetFilter, filter_func=filter_func)
            )
        raise TypeError(expr_or_filter_func)

    def update(self, source: Union[DatasetPatch, IDataset, Iterable[DatasetItem]]) -> Dataset:
        """
        Updates items of the current dataset from another dataset or an
        iterable (the source). Items from the source overwrite matching
        items in the current dataset. Unmatched items are just appended.

        If the source is a DatasetPatch, the removed items in the patch
        will be removed in the current dataset.

        If the source is a dataset, labels are matched. If the labels match,
        but the order is different, the annotation labels will be remapped to
        the current dataset label order during updating.

        Returns: self
        """

        self._data.update(source)
        return self

    def transform(self, method: Union[str, Type[Transform]], **kwargs) -> Dataset:
        """
        Applies some function to dataset items.

        Results are stored in-place. Modifications are applied lazily.
        Transforms are not allowed to change media type of dataset items.

        Args:
            method: The transformation to be applied to the dataset.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the dataset environment.
            **kwargs: Parameters for the transformation

        Returns: self
        """

        if isinstance(method, str):
            method = self.env.transforms[method]

        if not (inspect.isclass(method) and issubclass(method, Transform)):
            raise TypeError("Unexpected 'method' argument type: %s" % type(method))

        self._data.transform(method, **kwargs)
        if self.is_eager:
            self.init_cache()

        return self

    def run_model(
        self,
        model: Union[Launcher, Type[ModelTransform]],
        *,
        batch_size: int = 1,
        append_annotation: bool = False,
        num_workers: int = 0,
        **kwargs,
    ) -> Dataset:
        """
        Applies a model to dataset items' media and produces a dataset with
        media and annotations.

        Args:
            model: The model to be applied to the dataset
            batch_size: The number of dataset items processed
                simultaneously by the model
            append_annotation: Whether append new annotation to existed annotations
            num_workers: The number of worker threads to use for parallel inference.
                Set to 0 for single-process mode. Default is 0.
            **kwargs: Parameters for the model

        Returns: self
        """

        if isinstance(model, Launcher):
            return self.transform(
                ModelTransform,
                launcher=model,
                batch_size=batch_size,
                append_annotation=append_annotation,
                num_workers=num_workers,
                **kwargs,
            )
        elif inspect.isclass(model) and isinstance(model, ModelTransform):
            return self.transform(
                model,
                batch_size=batch_size,
                append_annotation=append_annotation,
                num_workers=num_workers,
                **kwargs,
            )
        else:
            raise TypeError("Unexpected 'model' argument type: %s" % type(model))

    def select(self, pred: Callable[[DatasetItem], bool]) -> Dataset:
        class _DatasetFilter(ItemTransform):
            def transform_item(self, item):
                if pred(item):
                    return item
                return None

        return self.transform(_DatasetFilter)

    @property
    def data_path(self) -> Optional[str]:
        return self._source_path

    @property
    def format(self) -> Optional[str]:
        return self._format

    @property
    def options(self) -> Dict[str, Any]:
        return self._options

    @property
    def is_modified(self) -> bool:
        return self._data.has_updated_items()

    def get_patch(self) -> DatasetPatch:
        return self._data.get_patch()

    @property
    def env(self) -> Environment:
        if self._env is None:
            return DEFAULT_ENVIRONMENT
        return self._env

    @property
    def is_cache_initialized(self) -> bool:
        return self._data.is_cache_initialized()

    @property
    def is_eager(self) -> bool:
        return self.eager if self.eager is not None else self._global_eager

    @property
    def is_bound(self) -> bool:
        return bool(self._source_path) and bool(self._format)

    def bind(
        self, path: str, format: Optional[str] = None, *, options: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Binds the dataset to a speific directory.
        Allows to set default saving parameters.

        The following saves will be done to this directory by default and will
        use the saved parameters.
        """

        self._source_path = path
        self._format = format or DEFAULT_FORMAT
        self._options = options or {}

    def flush_changes(self):
        self._data.flush_changes()

    @scoped
    def export(
        self,
        save_dir: str,
        format: Union[str, Type[Exporter]],
        *,
        progress_reporter: Optional[ProgressReporter] = None,
        error_policy: Optional[ExportErrorPolicy] = None,
        **kwargs,
    ) -> None:
        """
        Saves the dataset in some format.

        Args:
            save_dir: The output directory
            format: The desired output format.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the dataset environment.
            progress_reporter: An object to report progress
            error_policy: An object to report format-related errors
            **kwargs: Parameters for the format
        """

        if not save_dir:
            raise ValueError("Dataset export path is not specified")

        inplace = save_dir == self._source_path and format == self._format

        if isinstance(format, str):
            exporter = self.env.exporters[format]
        else:
            exporter = format

        if not (inspect.isclass(exporter) and issubclass(exporter, Exporter)):
            raise TypeError("Unexpected 'format' argument type: %s" % type(exporter))

        save_dir = osp.abspath(save_dir)
        if not osp.exists(save_dir):
            on_error_do(rmtree, save_dir, ignore_errors=True)
            inplace = False
        os.makedirs(save_dir, exist_ok=True)

        has_ctx_args = progress_reporter is not None or error_policy is not None

        if not progress_reporter:
            progress_reporter = NullProgressReporter()

        assert "ctx" not in kwargs
        exporter_kwargs = copy(kwargs)
        exporter_kwargs["stream"] = self._stream
        exporter_kwargs["ctx"] = ExportContext(
            progress_reporter=progress_reporter, error_policy=error_policy
        )

        try:
            if not inplace:
                try:
                    exporter.convert(self, save_dir=save_dir, **exporter_kwargs)
                except TypeError as e:
                    # TODO: for backward compatibility. To be removed after 0.3
                    if "unexpected keyword argument 'ctx'" not in str(e):
                        raise

                    if has_ctx_args:
                        warnings.warn(
                            f"It seems that '{format}' exporter "
                            "does not support progress and error reporting, "
                            "It will be disabled in datumaro==1.5.0.",
                            DeprecationWarning,
                        )
                    exporter_kwargs.pop("ctx")

                    exporter.convert(self, save_dir=save_dir, **exporter_kwargs)
            else:
                try:
                    exporter.patch(self, self.get_patch(), save_dir=save_dir, **exporter_kwargs)
                except TypeError as e:
                    # TODO: for backward compatibility. To be removed after 0.3
                    if "unexpected keyword argument 'ctx'" not in str(e):
                        raise

                    if has_ctx_args:
                        warnings.warn(
                            f"It seems that '{format}' exporter "
                            "does not support progress and error reporting, "
                            "It will be disabled in datumaro==1.5.0.",
                            DeprecationWarning,
                        )
                    exporter_kwargs.pop("ctx")

                    exporter.patch(self, self.get_patch(), save_dir=save_dir, **exporter_kwargs)
        except _ExportFail as e:
            raise e.__cause__

        self.bind(save_dir, format, options=copy(kwargs))
        if not self._stream:
            self.flush_changes()

    def save(self, save_dir: Optional[str] = None, **kwargs) -> None:
        options = dict(self._options)
        options.update(kwargs)

        self.export(save_dir or self._source_path, format=self._format, **options)

    @classmethod
    def load(cls, path: str, **kwargs) -> Dataset:
        return cls.import_from(path, format=DEFAULT_FORMAT, **kwargs)

    @classmethod
    def import_from(
        cls,
        path: str,
        format: Optional[str] = None,
        *,
        env: Optional[Environment] = None,
        progress_reporter: Optional[ProgressReporter] = None,
        error_policy: Optional[ImportErrorPolicy] = None,
        **kwargs,
    ) -> Dataset:
        """
        Creates a `Dataset` instance from a dataset on the disk.

        Args:
            path - The input file or directory path
            format - Dataset format.
                If a string is passed, it is treated as a plugin name,
                which is searched for in the `env` plugin context.
                If not set, will try to detect automatically,
                using the `env` plugin context.
            env - A plugin collection. If not set, the built-in plugins are used
            progress_reporter - An object to report progress.
                Implies earger loading.
            error_policy - An object to report format-related errors.
                Implies earger loading.
            **kwargs - Parameters for the format
        """

        if env is None:
            env = DEFAULT_ENVIRONMENT

        if not format:
            format = cls.detect(path, env=env)

        extractor_merger = None
        # TODO: remove importers, put this logic into extractors
        if format in env.importers:
            importer = env.make_importer(format)
            with logging_disabled(log.INFO):
                detected_sources = (
                    importer(path, stream=cls._stream, **kwargs)
                    if importer.can_stream
                    else importer(path, **kwargs)
                )
            extractor_merger = importer.get_extractor_merger()
        elif format in env.extractors:
            detected_sources = [{"url": path, "format": format, "options": kwargs}]
        else:
            raise UnknownFormatError(format)

        # TODO: probably, should not be available in lazy mode, because it
        # becomes unreliable and error-prone. For progress reporting it
        # makes little sense, because loading stage is spread over other
        # operations. Error reporting is going to be unreliable.
        has_ctx_args = progress_reporter is not None or error_policy is not None
        eager = has_ctx_args

        if not progress_reporter:
            progress_reporter = NullProgressReporter()
        pbars = progress_reporter.split(len(detected_sources))

        try:
            extractors = []
            for src_conf, pbar in zip(detected_sources, pbars):
                if not isinstance(src_conf, Source):
                    src_conf = Source(src_conf)

                extractor_kwargs = dict(src_conf.options)

                assert "ctx" not in extractor_kwargs
                extractor_kwargs["ctx"] = ImportContext(
                    progress_reporter=pbar, error_policy=error_policy
                )

                merge_policy = extractor_kwargs.get("merge_policy", DEFAULT_MERGE_POLICY)

                try:
                    extractors.append(
                        env.make_extractor(src_conf.format, src_conf.url, **extractor_kwargs)
                    )
                except TypeError as e:
                    # TODO: for backward compatibility. To be removed after 0.3
                    if "unexpected keyword argument 'ctx'" not in str(e):
                        raise

                    if has_ctx_args:
                        warnings.warn(
                            f"It seems that '{src_conf.format}' extractor "
                            "does not support progress and error reporting. "
                            "It will be disabled in datumaro==1.5.0.",
                            DeprecationWarning,
                        )
                    extractor_kwargs.pop("ctx")

                    extractors.append(
                        env.make_extractor(src_conf.format, src_conf.url, **extractor_kwargs)
                    )
            dataset = (
                cls(source=extractor_merger(extractors), env=env)
                if extractor_merger is not None
                else cls.from_extractors(*extractors, env=env, merge_policy=merge_policy)
            )
            if eager:
                dataset.init_cache()
        except _ImportFail as e:
            cause = e.__cause__ if getattr(e, "__cause__", None) is not None else e
            cause.__traceback__ = e.__traceback__
            raise DatasetImportError(f"Failed to import dataset '{format}' at '{path}'.") from cause
        except DatumaroError as e:
            cause = e.__cause__ if getattr(e, "__cause__", None) is not None else e
            cause.__traceback__ = e.__traceback__
            raise DatasetImportError(f"Failed to import dataset '{format}' at '{path}'.") from cause
        except Exception as e:
            raise DatasetImportError(f"Failed to import dataset '{format}' at '{path}'.") from e

        dataset._source_path = path
        dataset._format = format

        dataset = load_hash_key(path, dataset)
        return dataset

    @staticmethod
    def detect(path: str, *, env: Optional[Environment] = None, depth: int = 2) -> str:
        """
        Attempts to detect dataset format of a given directory.

        This function tries to detect a single format and fails if it's not
        possible. Check Environment.detect_dataset() for a function that
        reports status for each format checked.

        Args:
            path: The directory to check
            depth: The maximum depth for recursive search
            env: A plugin collection. If not set, the built-in plugins are used
        """

        if env is None:
            env = Environment()

        if depth < 0:
            raise ValueError("Depth cannot be less than zero")

        matches = env.detect_dataset(path, depth=depth)
        if not matches:
            raise NoMatchingFormatsError()
        elif 1 < len(matches):
            raise MultipleFormatsMatchError(matches)
        else:
            return matches[0]

    @property
    def is_stream(self) -> bool:
        return self._data.is_stream

    def clone(self) -> "Dataset":
        """Create a deep copy of this dataset.

        Returns:
            A cloned instance of the `Dataset`.
        """
        return deepcopy(self)

    def __getitem__(self, idx: int) -> DatasetItem:
        if not self._data.is_stream:
            return self._data[idx]
        raise StreamedItemError()


class StreamDataset(Dataset):
    _stream = True

    def __init__(
        self,
        source: Optional[IDataset] = None,
        *,
        infos: Optional[DatasetInfo] = None,
        categories: Optional[CategoriesInfo] = None,
        media_type: Optional[Type[MediaElement]] = None,
        ann_types: Optional[Set[AnnotationType]] = None,
        env: Optional[Environment] = None,
    ) -> None:
        assert env is None or isinstance(env, Environment), env
        self._env = env

        self._data = StreamDatasetStorage(
            source,
            infos=infos,
            categories=categories,
            media_type=media_type,
            ann_types=ann_types,
        )

        self._format = DEFAULT_FORMAT
        self._source_path = None
        self._options = {}

    @property
    def is_eager(self) -> bool:
        return False

    @classmethod
    def from_extractors(
        cls,
        *sources: IDataset,
        env: Optional[Environment] = None,
        merge_policy: str = DEFAULT_MERGE_POLICY,
    ) -> Dataset:
        """
        Creates a new dataset from one or several `Extractor`s.

        In case of a single input, creates a lazy wrapper around the input.
        In case of several inputs, unifies them and caches the resulting
        dataset. We cannot apply regular dataset merge, since items list cannot be accessed.

        Parameters:
            sources: one or many input extractors
            env: A context for plugins, which will be used for this dataset.
                If not specified, the builtin plugins will be used.
            merge_policy: Policy on how to merge multiple datasets.
                Possible options are "exact", "intersect", and "union".

        Returns:
            dataset: A new dataset with contents produced by input extractors
        """

        if len(sources) == 1:
            source = sources[0]
            dataset = cls(source=source, env=env)
        else:

            class _MergedStreamDataset(cls):
                def __init__(self, *sources: IDataset):
                    from datumaro.components.hl_ops import HLOps

                    self._merged = HLOps.merge(*sources, merge_policy=merge_policy)
                    self._data = self._merged._data
                    self._env = env
                    self._format = DEFAULT_FORMAT
                    self._source_path = None
                    self._options = {}

                def __iter__(self):
                    yield from self._merged

                @property
                def is_stream(self):
                    return True

                def subsets(self) -> Dict[str, DatasetSubset]:
                    return self._merged.subsets()

            return _MergedStreamDataset(*sources)

        return dataset


@contextmanager
def eager_mode(new_mode: bool = True, dataset: Optional[Dataset] = None) -> None:
    if dataset is not None:
        old_mode = dataset.eager

        try:
            dataset.eager = new_mode
            yield
        finally:
            dataset.eager = old_mode
    else:
        old_mode = Dataset._global_eager

        try:
            Dataset._global_eager = new_mode
            yield
        finally:
            Dataset._global_eager = old_mode



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from .base import SegmentAnythingBase
from .exporter import SegmentAnythingExporter
from .importer import SegmentAnythingImporter

__all__ = ["SegmentAnythingBase", "SegmentAnythingImporter", "SegmentAnythingExporter"]



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
import struct
from typing import Dict, Optional, Tuple

from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image, MediaElement, MediaType, PointCloud, Video, VideoFrame

from .common import Mapper, StringMapper


class MediaMapper(Mapper):
    @classmethod
    def forward(cls, obj: Optional[MediaElement]) -> bytes:
        if obj is None:
            return struct.pack("<I", MediaType.NONE)
        elif obj._type == MediaType.IMAGE:
            return ImageMapper.forward(obj)
        elif obj._type == MediaType.POINT_CLOUD:
            return PointCloudMapper.forward(obj)
        elif obj._type == MediaType.VIDEO:
            return VideoMapper.forward(obj)
        elif obj._type == MediaType.VIDEO_FRAME:
            return VideoFrameMapper.forward(obj)
        elif obj._type == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.forward(obj)
        else:
            raise DatumaroError(f"{obj._type} is not allowed for MediaMapper.")

    @classmethod
    def backward(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[Optional[MediaElement], int]:
        (media_type,) = struct.unpack_from("<I", _bytes, offset)

        if media_type == MediaType.NONE:
            return None, offset + 4
        elif media_type == MediaType.IMAGE:
            return ImageMapper.backward(_bytes, offset, media_path_prefix)
        elif media_type == MediaType.POINT_CLOUD:
            return PointCloudMapper.backward(_bytes, offset, media_path_prefix)
        elif media_type == MediaType.VIDEO:
            return VideoMapper.backward(_bytes, offset, media_path_prefix)
        elif media_type == MediaType.VIDEO_FRAME:
            return VideoFrameMapper.backward(_bytes, offset, media_path_prefix)
        elif media_type == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.backward(_bytes, offset, media_path_prefix)
        else:
            raise DatumaroError(f"{media_type} is not allowed for MediaMapper.")


class MediaElementMapper(Mapper):
    MAGIC_PATH = "/NOT/A/REAL/PATH"
    MEDIA_TYPE = MediaType.MEDIA_ELEMENT

    @classmethod
    def forward(cls, obj: MediaElement) -> bytes:
        bytes_arr = bytearray()
        bytes_arr.extend(struct.pack("<I", obj.type))
        path = getattr(obj, "path", cls.MAGIC_PATH)
        bytes_arr.extend(StringMapper.forward(path))

        return bytes(bytes_arr)

    @classmethod
    def backward_dict(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[Dict, int]:
        (media_type,) = struct.unpack_from("<I", _bytes, offset)
        assert media_type == cls.MEDIA_TYPE, f"Expect {cls.MEDIA_TYPE} but actual is {media_type}."
        offset += 4
        path, offset = StringMapper.backward(_bytes, offset)
        if path == cls.MAGIC_PATH:
            path = None
        return {
            "type": media_type,
            "path": path
            if path == cls.MAGIC_PATH or media_path_prefix is None
            else osp.join(media_path_prefix[cls.MEDIA_TYPE], path),
        }, offset

    @classmethod
    def backward(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[MediaElement, int]:
        _, offset = cls.backward_dict(_bytes, offset, media_path_prefix)
        return MediaElement(), offset


class ImageMapper(MediaElementMapper):
    MAGIC_SIZE_FOR_NONE = (-1583, -1597)
    MEDIA_TYPE = MediaType.IMAGE

    @classmethod
    def forward(cls, obj: Image) -> bytes:
        size = obj.size if obj.has_size else cls.MAGIC_SIZE_FOR_NONE

        bytes_arr = bytearray()
        bytes_arr.extend(super().forward(obj))
        bytes_arr.extend(struct.pack("<ii", size[0], size[1]))

        return bytes(bytes_arr)

    @classmethod
    def backward(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[Image, int]:
        media_dict, offset = cls.backward_dict(_bytes, offset, media_path_prefix)
        height, width = struct.unpack_from("<ii", _bytes, offset)
        size = (height, width)
        offset += 8
        return (
            Image.from_file(
                path=media_dict["path"], size=size if size != cls.MAGIC_SIZE_FOR_NONE else None
            ),
            offset,
        )


class VideoMapper(MediaElementMapper):
    MAGIC_END_FRAME_FOR_NONE = 4294967295  # max value of unsigned int32
    MEDIA_TYPE = MediaType.VIDEO

    @classmethod
    def forward(cls, obj: Video) -> bytes:
        end_frame = obj._end_frame if obj._end_frame else cls.MAGIC_END_FRAME_FOR_NONE

        bytes_arr = bytearray()
        bytes_arr.extend(super().forward(obj))
        bytes_arr.extend(struct.pack("<III", obj._step, obj._start_frame, end_frame))

        return bytes(bytes_arr)

    @classmethod
    def backward(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[Video, int]:
        media_dict, offset = cls.backward_dict(_bytes, offset, media_path_prefix)
        step, start_frame, end_frame = struct.unpack_from("<III", _bytes, offset)
        offset += 12
        video = Video(
            path=media_dict["path"],
            step=step,
            start_frame=start_frame,
            end_frame=end_frame if end_frame != cls.MAGIC_END_FRAME_FOR_NONE else None,
        )
        return (video, offset)


class VideoFrameMapper(MediaElementMapper):
    MEDIA_TYPE = MediaType.VIDEO_FRAME

    @classmethod
    def forward(cls, obj: VideoFrame) -> bytes:
        bytes_arr = bytearray()
        bytes_arr.extend(super().forward(obj))
        bytes_arr.extend(struct.pack("<I", obj.index))

        return bytes(bytes_arr)

    @classmethod
    def backward(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[VideoFrame, int]:
        media_dict, offset = cls.backward_dict(_bytes, offset, media_path_prefix)
        (frame_index,) = struct.unpack_from("<I", _bytes, offset)
        video = Video(media_dict["path"])
        offset += 4
        return (
            VideoFrame(video, frame_index),
            offset,
        )


class PointCloudMapper(MediaElementMapper):
    MEDIA_TYPE = MediaType.POINT_CLOUD

    @classmethod
    def forward(cls, obj: PointCloud) -> bytes:
        bytes_arr = bytearray()
        bytes_arr.extend(super().forward(obj))
        bytes_arr.extend(struct.pack("<I", len(obj.extra_images)))
        for img in obj.extra_images:
            bytes_arr.extend(ImageMapper.forward(img))

        return bytes(bytes_arr)

    @classmethod
    def backward(
        cls,
        _bytes: bytes,
        offset: int = 0,
        media_path_prefix: Optional[Dict[MediaType, str]] = None,
    ) -> Tuple[PointCloud, int]:
        media_dict, offset = cls.backward_dict(_bytes, offset, media_path_prefix)
        (len_extra_images,) = struct.unpack_from("<I", _bytes, offset)
        offset += 4

        extra_images = []
        for _ in range(len_extra_images):
            img, offset = ImageMapper.backward(_bytes, offset, media_path_prefix)
            extra_images.append(img)

        return PointCloud.from_file(path=media_dict["path"], extra_images=extra_images), offset



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import os.path as osp
import re
from collections import OrderedDict
from typing import List, Optional

import numpy as np

from datumaro.components.annotation import AnnotationType, Bbox, LabelCategories, Mask
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.lazy_plugin import extra_deps
from datumaro.components.media import Image
from datumaro.util.image import decode_image, lazy_image
from datumaro.util.tf_util import has_feature
from datumaro.util.tf_util import import_tf as _import_tf

from .format import DetectionApiPath, TfrecordImporterType

tf = _import_tf()


def clamp(value, _min, _max):
    return max(min(_max, value), _min)


@extra_deps("tensorflow")
class TfDetectionApiBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        tfrecord_importer_type: TfrecordImporterType = TfrecordImporterType.default,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        assert osp.isfile(path), path
        images_dir = ""
        root_dir = osp.dirname(osp.abspath(path))
        if osp.basename(root_dir) == DetectionApiPath.ANNOTATIONS_DIR:
            root_dir = osp.dirname(root_dir)
            images_dir = osp.join(root_dir, DetectionApiPath.IMAGES_DIR)
            if not osp.isdir(images_dir):
                images_dir = ""

        if not subset:
            subset = osp.splitext(osp.basename(path))[0]
        super().__init__(subset=subset, ctx=ctx)

        self._features = {
            "image/filename": tf.io.FixedLenFeature([], tf.string),
            "image/source_id": tf.io.FixedLenFeature([], tf.string),
            "image/height": tf.io.FixedLenFeature([], tf.int64),
            "image/width": tf.io.FixedLenFeature([], tf.int64),
            "image/encoded": tf.io.FixedLenFeature([], tf.string),
            "image/format": tf.io.FixedLenFeature([], tf.string),
            # use varlen to avoid errors when this field is missing
            "image/key/sha256": tf.io.VarLenFeature(tf.string),
            # Object boxes and classes.
            "image/object/bbox/xmin": tf.io.VarLenFeature(tf.float32),
            "image/object/bbox/xmax": tf.io.VarLenFeature(tf.float32),
            "image/object/bbox/ymin": tf.io.VarLenFeature(tf.float32),
            "image/object/bbox/ymax": tf.io.VarLenFeature(tf.float32),
            "image/object/class/label": tf.io.VarLenFeature(tf.int64),
            "image/object/class/text": tf.io.VarLenFeature(tf.string),
            "image/object/mask": tf.io.VarLenFeature(tf.string),
        }
        if tfrecord_importer_type == TfrecordImporterType.roboflow:
            del self._features["image/source_id"]

        items, labels = self._parse_tfrecord_file(path, self._subset, images_dir)
        self._categories = self._load_categories(labels)
        self._items = items

    @staticmethod
    def _load_categories(labels):
        label_categories = LabelCategories().from_iterable(
            e[0] for e in sorted(labels.items(), key=lambda item: item[1])
        )
        return {AnnotationType.label: label_categories}

    @staticmethod
    def _parse_labelmap(text):
        id_pattern = r"(?:id\s*:\s*(?P<id>\d+))"
        name_pattern = r"(?:name\s*:\s*[\'\"](?P<name>.*?)[\'\"])"
        entry_pattern = r"(\{(?:[\s\n]*(?:%(id)s|%(name)s)[\s\n]*){2}\})+" % {
            "id": id_pattern,
            "name": name_pattern,
        }
        matches = re.finditer(entry_pattern, text)

        labelmap = {}
        for match in matches:
            label_id = match.group("id")
            label_name = match.group("name")
            if label_id is not None and label_name is not None:
                labelmap[label_name] = int(label_id)

        return labelmap

    def _parse_tfrecord_file(self, filepath, subset, images_dir):
        dataset = tf.data.TFRecordDataset(filepath)

        files = os.listdir(osp.dirname(filepath))
        for filename in files:
            if DetectionApiPath.LABELMAP_FILE in filename:
                labelmap_path = osp.join(osp.dirname(filepath), filename)
                break

        dataset_labels = OrderedDict()
        if osp.exists(labelmap_path):
            with open(labelmap_path, "r", encoding="utf-8") as f:
                labelmap_text = f.read()
            dataset_labels.update(
                {label: id - 1 for label, id in self._parse_labelmap(labelmap_text).items()}
            )

        dataset_items = []

        for record in dataset:
            parsed_record = tf.io.parse_single_example(record, self._features)
            frame_id = parsed_record.get("image/source_id", None)
            frame_id = frame_id.numpy().decode("utf-8") if frame_id else frame_id
            frame_filename = parsed_record.get("image/filename", None).numpy().decode("utf-8")
            frame_height = tf.cast(parsed_record.get("image/height", 0), tf.int64).numpy().item()
            frame_width = tf.cast(parsed_record.get("image/width", 0), tf.int64).numpy().item()
            frame_image = parsed_record["image/encoded"].numpy()
            xmins = tf.sparse.to_dense(parsed_record["image/object/bbox/xmin"]).numpy()
            ymins = tf.sparse.to_dense(parsed_record["image/object/bbox/ymin"]).numpy()
            xmaxs = tf.sparse.to_dense(parsed_record["image/object/bbox/xmax"]).numpy()
            ymaxs = tf.sparse.to_dense(parsed_record["image/object/bbox/ymax"]).numpy()
            label_ids = tf.sparse.to_dense(parsed_record["image/object/class/label"]).numpy()
            labels = tf.sparse.to_dense(
                parsed_record["image/object/class/text"], default_value=b""
            ).numpy()
            masks = tf.sparse.to_dense(
                parsed_record["image/object/mask"], default_value=b""
            ).numpy()

            for label, label_id in zip(labels, label_ids):
                label = label.decode("utf-8")
                if not label:
                    continue
                if label_id <= 0:
                    continue
                if label in dataset_labels:
                    continue
                dataset_labels[label] = label_id - 1

            item_id = osp.splitext(frame_filename)[0]

            annotations = []
            for shape_id, shape in enumerate(np.dstack((labels, xmins, ymins, xmaxs, ymaxs))[0]):
                label = shape[0].decode("utf-8")

                mask = None
                if len(masks) != 0:
                    mask = masks[shape_id]

                if mask is not None:
                    if isinstance(mask, bytes):
                        mask = lazy_image(mask, decode_image)
                    annotations.append(Mask(image=mask, label=dataset_labels.get(label)))
                else:
                    x = clamp(shape[1] * frame_width, 0, frame_width)
                    y = clamp(shape[2] * frame_height, 0, frame_height)
                    w = clamp(shape[3] * frame_width, 0, frame_width) - x
                    h = clamp(shape[4] * frame_height, 0, frame_height) - y
                    annotations.append(Bbox(x, y, w, h, label=dataset_labels.get(label)))

            image_size = None
            if frame_height and frame_width:
                image_size = (frame_height, frame_width)

            image = None
            if frame_image:
                if isinstance(frame_image, np.ndarray):
                    image = Image.from_numpy(data=frame_image, size=image_size)
                else:
                    image = Image.from_bytes(data=frame_image, size=image_size)
            elif frame_filename:
                image = Image.from_file(path=osp.join(images_dir, frame_filename), size=image_size)

            for ann in annotations:
                self._ann_types.add(ann.type)

            dataset_items.append(
                DatasetItem(
                    id=item_id,
                    subset=subset,
                    media=image,
                    annotations=annotations,
                    attributes={"source_id": frame_id},
                )
            )

        return dataset_items, dataset_labels


@extra_deps("tensorflow")
class TfDetectionApiImporter(Importer):
    _FORMAT_EXT = ".tfrecord"

    @classmethod
    def find_sources(cls, path):
        sources = cls._find_sources_recursive(
            path=path,
            ext=cls._FORMAT_EXT,
            extractor_name="tf_detection_api",
        )
        if len(sources) == 0:
            return []

        desired_feature = {
            "image/source_id": tf.io.FixedLenFeature([], tf.string),
        }

        subsets = {}
        for source in sources:
            if has_feature(path=source["url"], feature=desired_feature):
                subset_name = osp.basename(source["url"]).split(".")[-2]
                subsets[subset_name] = source["url"]

        sources = [
            {
                "url": url,
                "format": "tf_detection_api",
            }
            for _, url in subsets.items()
        ]

        return sources

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._FORMAT_EXT]



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

# pylint: disable=exec-used


import logging as log
import os.path as osp
import shutil
import urllib
from dataclasses import dataclass, fields
from typing import Dict, List, Optional

import numpy as np
from openvino.runtime import Core
from tqdm import tqdm

from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.launcher import LauncherWithModelInterpreter
from datumaro.errors import DatumaroError
from datumaro.util.definitions import get_datumaro_cache_dir
from datumaro.util.samples import get_samples_path


class _OpenvinoImporter(CliPlugin):
    @staticmethod
    def _parse_output_layers(s):
        return [s.strip() for s in s.split(",")]

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-d", "--description", required=True, help="Path to the model description file (.xml)"
        )
        parser.add_argument(
            "-w", "--weights", required=True, help="Path to the model weights file (.bin)"
        )
        parser.add_argument(
            "-i",
            "--interpreter",
            required=True,
            help="Path to the network output interprter script (.py)",
        )
        parser.add_argument("--device", default="CPU", help="Target device (default: %(default)s)")
        parser.add_argument(
            "--output-layers",
            type=cls._parse_output_layers,
            help="A comma-separated list of extra output layers",
        )
        return parser

    @staticmethod
    def copy_model(model_dir, model):
        shutil.copy(model["description"], osp.join(model_dir, osp.basename(model["description"])))
        model["description"] = osp.basename(model["description"])

        shutil.copy(model["weights"], osp.join(model_dir, osp.basename(model["weights"])))
        model["weights"] = osp.basename(model["weights"])

        shutil.copy(model["interpreter"], osp.join(model_dir, osp.basename(model["interpreter"])))
        model["interpreter"] = osp.basename(model["interpreter"])


@dataclass
class OpenvinoModelInfo:
    interpreter: Optional[str]
    description: Optional[str]
    weights: Optional[str]
    model_dir: Optional[str]

    def validate(self):
        """Validate integrity of the member variables"""

        def _validate(key: str):
            path = getattr(self, key)
            if not osp.isfile(path):
                path = osp.join(self.model_dir, path)
            if not osp.isfile(path):
                raise DatumaroError(f'Failed to open model {key} file "{path}"')
            setattr(self, key, path)

        for field in fields(self):
            if field.name != "model_dir":
                _validate(field.name)


@dataclass
class BuiltinOpenvinoModelInfo(OpenvinoModelInfo):
    downloadable_models = {
        "clip_text_ViT-B_32",
        "clip_visual_ViT-B_32",
        "googlenet-v4-tf",
    }

    @classmethod
    def create_from_model_name(cls, model_name: str) -> "BuiltinOpenvinoModelInfo":
        openvino_plugin_samples_dir = get_samples_path()
        interpreter = osp.join(openvino_plugin_samples_dir, model_name + "_interp.py")
        interpreter = interpreter if osp.exists(interpreter) else interpreter

        model_dir = get_datumaro_cache_dir()

        # Please visit open-model-zoo repository for OpenVINO public models if you are interested in
        # https://github.com/openvinotoolkit/open_model_zoo/blob/master/models/public/index.md
        url_folder = "https://storage.openvinotoolkit.org/repositories/datumaro/models/"

        description = osp.join(model_dir, model_name + ".xml")
        if not osp.exists(description):
            description = (
                cls._download_file(osp.join(url_folder, model_name + ".xml"), description)
                if model_name in cls.downloadable_models
                else None
            )

        weights = osp.join(model_dir, model_name + ".bin")
        if not osp.exists(weights):
            weights = (
                cls._download_file(osp.join(url_folder, model_name + ".bin"), weights)
                if model_name in cls.downloadable_models
                else None
            )

        return cls(
            interpreter=interpreter,
            description=description,
            weights=weights,
            model_dir=model_dir,
        )

    @staticmethod
    def _download_file(url: str, file_root: str) -> str:
        log.info('Downloading: "{}" to {}\n'.format(url, file_root))
        req = urllib.request.Request(url)
        with urllib.request.urlopen(req) as source, open(file_root, "wb") as output:  # nosec B310
            with tqdm(
                total=int(source.info().get("Content-Length")),
                ncols=80,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
            ) as loop:
                while True:
                    buffer = source.read(8192)
                    if not buffer:
                        break

                    output.write(buffer)
                    loop.update(len(buffer))
        return file_root

    def override(self, other: OpenvinoModelInfo) -> None:
        """Override builtin model variables to other"""

        def _apply(key: str) -> None:
            other_item = getattr(other, key)
            self_item = getattr(self, key)
            if other_item is None and self_item:
                log.info(f"Override description with the builtin model {key}: {self.description}.")
                setattr(other, key, self_item)

        for field in fields(self):
            _apply(field.name)


class OpenvinoLauncher(LauncherWithModelInterpreter):
    cli_plugin = _OpenvinoImporter

    def __init__(
        self,
        description: Optional[str] = None,
        weights: Optional[str] = None,
        interpreter: Optional[str] = None,
        model_dir: Optional[str] = None,
        model_name: Optional[str] = None,
        output_layers: List[str] = [],
        device: Optional[str] = None,
        compile_model_config: Optional[Dict] = None,
    ):
        model_info = OpenvinoModelInfo(
            interpreter=interpreter,
            description=description,
            weights=weights,
            model_dir=model_dir,
        )
        if model_name:
            builtin_model_info = BuiltinOpenvinoModelInfo.create_from_model_name(model_name)
            builtin_model_info.override(model_info)

        model_info.validate()

        super().__init__(model_interpreter_path=model_info.interpreter)

        self.model_info = model_info

        self._device = device or "CPU"
        self._compile_model_config = compile_model_config

        self._core = Core()
        self._network = self._core.read_model(model_info.description, model_info.weights)

        if output_layers:
            log.info(f"Add additional output layers {output_layers} to the model outputs.")
            self._network.add_outputs(output_layers)

        self._check_model_support(self._network, self._device)
        self._load_executable_net()

    @property
    def inputs(self):
        return self._network.inputs

    @property
    def outputs(self):
        return self._network.outputs

    def _check_model_support(self, net, device):
        not_supported_layers = set(
            name for name, dev in self._core.query_model(net, device).items() if not dev
        )
        if len(not_supported_layers) != 0:
            log.error(
                "The following layers are not supported "
                "by the plugin for device '%s': %s." % (device, ", ".join(not_supported_layers))
            )
            raise NotImplementedError("Some layers are not supported on the device")

    def _load_executable_net(self, batch_size: int = 1):
        network = self._network

        iter_inputs = iter(network.inputs)
        self._input_blob = next(iter_inputs)

        is_dynamic_layout = False
        try:
            self._input_layout = self._input_blob.shape
        except ValueError:
            # In case of that the input has dynamic shape
            self._input_layout = self._input_blob.partial_shape
            is_dynamic_layout = True

        if is_dynamic_layout:
            self._input_layout[0] = batch_size
            network.reshape({self._input_blob: self._input_layout})
        else:
            model_batch_size = self._input_layout[0]
            if batch_size != model_batch_size:
                log.warning(
                    "Input layout of the model is static, so that we cannot change "
                    f"the model batch size ({model_batch_size}) to batch size ({batch_size})! "
                    "Set the batch size to {model_batch_size}."
                )
                batch_size = model_batch_size

        self._batch_size = batch_size

        self._net = self._core.compile_model(
            model=network,
            device_name=self._device,
            config=self._compile_model_config,
        )
        self._request = self._net.create_infer_request()

    def infer(self, inputs: LauncherInputType) -> List[ModelPred]:
        batch_size = len(inputs)
        if self._batch_size < batch_size:
            self._load_executable_net(batch_size)

        inputs = (
            {self._input_blob.get_any_name(): inputs} if isinstance(inputs, np.ndarray) else inputs
        )
        results = self._request.infer(inputs=inputs)

        outputs_group_by_item = [
            {key.any_name: output for key, output in zip(results.keys(), outputs)}
            for outputs in zip(*results.values())
        ]

        return outputs_group_by_item



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT


from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import pyarrow as pa

from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image, MediaElement, MediaType, PointCloud
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import Mapper
from datumaro.util.image import decode_image, encode_image, load_image


class MediaMapper(Mapper):
    @classmethod
    def forward(cls, obj: Optional[MediaElement], **options) -> Dict[str, Any]:
        if obj is None:
            return {"type": int(MediaType.NONE)}
        if obj._type == MediaType.IMAGE:
            return ImageMapper.forward(obj, **options)
        if obj._type == MediaType.POINT_CLOUD:
            return PointCloudMapper.forward(obj, **options)
        if obj._type == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.forward(obj)
        raise DatumaroError(f"{obj._type} is not allowed for MediaMapper.")

    @classmethod
    def backward(
        cls,
        media_struct: pa.StructScalar,
        idx: int,
        table: pa.Table,
        table_path: str,
    ) -> Optional[MediaElement]:
        media_type = MediaType(media_struct.get("type").as_py())

        if media_type == MediaType.NONE:
            return None
        if media_type == MediaType.IMAGE:
            return ImageMapper.backward(media_struct, idx, table, table_path)
        if media_type == MediaType.POINT_CLOUD:
            return PointCloudMapper.backward(media_struct, idx, table, table_path)
        if media_type == MediaType.MEDIA_ELEMENT:
            return MediaElementMapper.backward(media_struct, idx, table, table_path)
        raise DatumaroError(f"{media_type} is not allowed for MediaMapper.")


class MediaElementMapper(Mapper):
    MAGIC_PATH = "/NOT/A/REAL/PATH"
    MEDIA_TYPE = MediaType.MEDIA_ELEMENT

    @classmethod
    def forward(cls, obj: MediaElement) -> Dict[str, Any]:
        return {"type": int(obj.type)}

    @classmethod
    def backward(
        cls,
        media_dict: Dict[str, Any],
        idx: int,
        table: pa.Table,
        table_path: str,
    ) -> MediaElement:
        return MediaElement()


class ImageMapper(MediaElementMapper):
    MEDIA_TYPE = MediaType.IMAGE
    AVAILABLE_SCHEMES = ("AS-IS", "PNG", "TIFF", "JPEG/95", "JPEG/75", "NONE")

    @classmethod
    def encode(cls, obj: Image, scheme: str = "PNG") -> Optional[bytes]:
        if scheme is None or scheme == "NONE":
            return None
        if scheme == "AS-IS":
            _bytes = obj.bytes
            if _bytes is not None:
                return _bytes
            # try to encode in PNG
            scheme = "PNG"

        options = {}
        if scheme.startswith("JPEG"):
            quality = int(scheme.split("/")[-1])
            options["ext"] = "JPEG"
            options["jpeg_quality"] = quality
        elif scheme == "PNG":
            options["ext"] = "PNG"
        elif scheme == "TIFF":
            options["ext"] = "TIFF"
        else:
            raise NotImplementedError

        data = obj.data
        if data is not None:
            return encode_image(obj.data, **options)
        return None

    @classmethod
    def decode(
        cls, path: Optional[str] = None, data: Optional[bytes] = None
    ) -> Optional[np.ndarray]:
        if path is None and data is None:
            return None
        if data is not None:
            return decode_image(data, np.uint8)
        if path is not None:
            return load_image(path, np.uint8)

    @classmethod
    def forward(
        cls, obj: Image, encoder: Union[str, Callable[[Image], bytes]] = "PNG"
    ) -> Dict[str, Any]:
        out = super().forward(obj)

        _bytes = encoder(obj) if isinstance(encoder, Callable) else cls.encode(obj, scheme=encoder)

        path = None if _bytes is not None else getattr(obj, "path", None)

        out["image"] = {
            "has_bytes": _bytes is not None,
            "bytes": _bytes,
            "path": path,
            "size": obj.size,
        }
        return out

    @classmethod
    def backward(
        cls,
        media_struct: pa.StructScalar,
        idx: int,
        table: pa.Table,
        table_path: str,
    ) -> Image:
        image_struct = media_struct.get("image")

        if path := image_struct.get("path").as_py():
            return Image.from_file(
                path=path,
                size=image_struct.get("size").as_py(),
            )

        return Image.from_bytes(
            data=lambda: pa.ipc.open_file(pa.memory_map(table_path, "r"))
            .read_all()
            .column("media")[idx]
            .get("image")
            .get("bytes")
            .as_py(),
            size=image_struct.get("size").as_py(),
        )

    @classmethod
    def backward_extra_image(
        cls, image_struct: pa.StructScalar, idx: int, table: pa.Table, extra_image_idx: int
    ) -> Image:
        if path := image_struct.get("path").as_py():
            return Image.from_file(
                path=path,
                size=image_struct.get("size").as_py(),
            )

        return Image.from_bytes(
            data=lambda: table.column("media")[idx]
            .get("point_cloud")
            .get("extra_images")[extra_image_idx]
            .get("bytes")
            .as_py(),
            size=image_struct.get("size").as_py(),
        )


# TODO: share binary for extra images
class PointCloudMapper(MediaElementMapper):
    MEDIA_TYPE = MediaType.POINT_CLOUD
    B64_PREFIX = "//B64_ENCODED//"

    @classmethod
    def forward(
        cls, obj: PointCloud, encoder: Union[str, Callable[[PointCloud], bytes]] = "PNG"
    ) -> Dict[str, Any]:
        out = super().forward(obj)

        if isinstance(encoder, Callable):
            _bytes = encoder(obj)
        elif encoder != "NONE":
            _bytes = obj.data
        else:
            _bytes = None

        path = None if _bytes is not None else getattr(obj, "path", None)

        out["point_cloud"] = {
            "has_bytes": _bytes is not None,
            "bytes": _bytes,
            "path": path,
            "extra_images": [
                ImageMapper.forward(img, encoder=encoder)["image"] for img in obj.extra_images
            ],
        }

        return out

    @classmethod
    def backward(
        cls,
        media_struct: pa.StructScalar,
        idx: int,
        table: pa.Table,
        table_path: str,
    ) -> PointCloud:
        point_cloud_struct = media_struct.get("point_cloud")

        extra_images = [
            ImageMapper.backward_extra_image(image_struct, idx, table, extra_image_idx)
            for extra_image_idx, image_struct in enumerate(point_cloud_struct.get("extra_images"))
        ]

        if path := point_cloud_struct.get("path").as_py():
            return PointCloud.from_file(path=path, extra_images=extra_images)

        return PointCloud.from_bytes(
            data=pa.ipc.open_file(pa.memory_map(table_path, "r"))
            .read_all()
            .column("media")[idx]
            .get("point_cloud")
            .get("bytes")
            .as_py(),
            extra_images=extra_images,
        )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Tuple

import cv2
import numpy as np

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation, FeatureVector
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image
from datumaro.plugins.openvino_plugin.samples.utils import rescale_img_keeping_aspect_ratio


class SAMEncoderInterpreter(IModelInterpreter):
    h_model = 1024
    w_model = 1024

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data
        output = rescale_img_keeping_aspect_ratio(img, self.h_model, self.w_model, padding=False)

        # From BGR to RGB
        output.image = cv2.cvtColor(output.image, cv2.COLOR_BGR2RGB)
        # From HWC to CHW
        output.image = output.image.transpose(2, 0, 1)
        # To FP32
        output.image = output.image.astype(np.float32)

        return output.image, output.scale

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        embed = pred.get("image_embeddings")

        assert embed is not None, 'Model output should have "image_embeddings".'

        return [FeatureVector(vector=embed)]

    def get_categories(self):
        return None



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from .base import ArrowBase
from .exporter import ArrowExporter
from .importer import ArrowImporter

__all__ = ["ArrowBase", "ArrowExporter", "ArrowImporter"]



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.dataset import Dataset
from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.util.os_util import make_file_name

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import FilterModes, generate_next_file_name


def build_parser(parser_ctor=argparse.ArgumentParser):
    builtin_readers = sorted(
        set(DEFAULT_ENVIRONMENT.importers) | set(DEFAULT_ENVIRONMENT.extractors)
    )
    builtin_writers = sorted(DEFAULT_ENVIRONMENT.exporters)

    parser = parser_ctor(
        help="Convert an existing dataset to another format",
        description="""
        Converts a dataset from one format to another.
        You can add your own formats and do many more by creating a
        Datumaro project.|n
        |n
        This command serves as an alias for the "create", "import", and
        "export" commands, allowing to obtain the same results simpler
        and faster. Check descriptions of these commands for more info.|n
        |n
        Supported input formats: {}|n
        |n
        Supported output formats: {}|n
        |n
        Examples:|n
        - Export a dataset as a PASCAL VOC dataset, include images:|n
        |s|s%(prog)s -i src/path -f voc -- --save-media|n
        |n
        - Export a dataset as a COCO dataset to a specific directory:|n
        |s|s%(prog)s -i src/path -f coco -o path/I/like/
        """.format(
            ", ".join(builtin_readers), ", ".join(builtin_writers)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "-i",
        "--input-path",
        default=".",
        dest="source",
        help="Input dataset path (default: current dir)",
    )
    parser.add_argument(
        "-if", "--input-format", help="Input dataset format. Will try to detect, if not specified."
    )
    parser.add_argument("-f", "--output-format", required=True, help="Output format")
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: a subdir in the current one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-e",
        "--filter",
        help='XML XPath filter expression for dataset items. Read "filter" '
        "command docs for more info",
    )
    parser.add_argument(
        "--filter-mode",
        default=FilterModes.i.name,
        type=FilterModes.parse,
        help="Filter mode, one of %s (default: %s)"
        % (", ".join(FilterModes.list_options()), "%(default)s"),
    )
    parser.add_argument(
        "--encryption-key",
        help="Secret key. It is required only if the input dataset is encrypted.",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for output format (pass '-- -h' for help). "
        "Must be specified after the main command arguments",
    )
    parser.set_defaults(command=convert_command)

    return parser


def get_sensitive_args():
    return {
        convert_command: ["source", "dst_dir", "extra_args"],
    }


def convert_command(args):
    env = DEFAULT_ENVIRONMENT

    try:
        exporter = env.exporters[args.output_format]
    except KeyError:
        raise CliException("Exporter for format '%s' is not found" % args.output_format)
    extra_args = exporter.parse_cmdline(args.extra_args)

    filter_args = FilterModes.make_filter_args(args.filter_mode)

    fmt = args.input_format
    if not args.input_format:
        matches = env.detect_dataset(args.source)
        if len(matches) == 0:
            log.error(
                "Failed to detect dataset format. "
                "Try to specify format with '-if/--input-format' parameter."
            )
            return 1
        elif len(matches) != 1:
            log.error(
                "Multiple formats match the dataset: %s. "
                "Try to specify format with '-if/--input-format' parameter.",
                ", ".join(matches),
            )
            return 2

        fmt = matches[0]
        log.info(f"Source dataset format detected as {fmt}")

    if fmt == args.output_format:
        log.error(f"The source data format and the output data format is same as {fmt}.")
        return 3

    source = osp.abspath(args.source)

    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name(
            "%s-%s" % (osp.basename(source), make_file_name(args.output_format))
        )
    dst_dir = osp.abspath(dst_dir)

    import_kwargs = {}
    if args.encryption_key:
        import_kwargs["encryption_key"] = args.encryption_key
    dataset = Dataset.import_from(source, fmt, **import_kwargs)

    log.info("Exporting the dataset")
    if args.filter:
        dataset = dataset.filter(args.filter, **filter_args)
    dataset.export(format=args.output_format, save_dir=dst_dir, **extra_args)

    log.info("Dataset exported to '%s' as '%s'" % (dst_dir, args.output_format))

    return 0



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.project import ProjectBuildTargets
from datumaro.util import str_to_bool
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    builtins = sorted(DEFAULT_ENVIRONMENT.transforms)

    parser = parser_ctor(
        help="Transform project",
        description="""
        Applies a batch operation to a dataset and produces a new dataset.|n
        |n
        By default, datasets are updated in-place. The '-o/--output-dir'
        option can be used to specify another output directory. When
        updating in-place, use the '--overwrite' parameter (in-place
        updates fail by default to prevent data loss), unless a project
        target is modified.|n
        |n
        Builtin transforms: {}|n
        |n
        This command has the following invocation syntax:
        - %(prog)s <target dataset revpath>|n
        |n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        The current project (-p/--project) is also used as a context for
        plugins, so it can be useful for dataset paths having custom formats.
        When not specified, the current project's working tree is used.|n
        |n
        The command can be applied to a dataset or a project build target,
        a stage or the combined 'project' target, in which case all the
        targets will be affected. A build tree stage will be recorded
        if '--stage' is enabled, and the resulting dataset(-s) will be
        saved if '--apply' is enabled.|n
        |n
        Examples:|n
        - Convert instance polygons to masks:|n |n
        |s|s%(prog)s -t polygons_to_masks|n
        |n
        - Rename dataset items by a regular expression:|n |n
        |s|s- Replace 'pattern' with 'replacement':|n |n
        |s|s|s|s%(prog)s -t rename -- -e '|pattern|replacement|'|n
        |n
        |s|s- Remove 'frame_' from item ids:|n |n
        |s|s|s|s%(prog)s -t rename -- -e '|frame_(\\d+)|\\1|'|n
        |n
        - Split a dataset randomly:|n |n
        |s|s%(prog)s -t random_split --overwrite path/to/dataset:voc
        """.format(
            ", ".join(builtins)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "_positionals", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )  # workaround for -- eaten by positionals
    parser.add_argument(
        "target", nargs="?", default="project", help="Target dataset revpath (default: project)"
    )
    parser.add_argument(
        "-t", "--transform", required=True, help="Transform to apply to the dataset"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="""
            Output directory. Can be omitted for main project targets
            (i.e. data sources and the 'project' target, but not
            intermediate stages) and dataset targets.
            If not specified, the results will be saved inplace.
            """,
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.add_argument(
        "--stage",
        type=str_to_bool,
        default=True,
        help="""
            Include this action as a project build step.
            If true, this operation will be saved in the project
            build tree, allowing to reproduce the resulting dataset later.
            Applicable only to main project targets (i.e. data sources
            and the 'project' target, but not intermediate stages)
            (default: %(default)s)
            """,
    )
    parser.add_argument(
        "--apply",
        type=str_to_bool,
        default=True,
        help="Run this command immediately. If disabled, only the "
        "build tree stage will be written (default: %(default)s)",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for transformation (pass '-- -h' for help). "
        "Must be specified after the main command arguments and after "
        "the '--' separator",
    )
    parser.set_defaults(command=transform_command)

    return parser


def get_sensitive_args():
    return {
        transform_command: ["dst_dir", "project_dir", "extra_args", "target"],
    }


@scoped
def transform_command(args):
    has_sep = "--" in args._positionals
    if has_sep:
        pos = args._positionals.index("--")
        if 1 < pos:
            raise argparse.ArgumentError(None, message="Expected no more than 1 target argument")
    else:
        pos = 1
    args.target = (args._positionals[:pos] or [ProjectBuildTargets.MAIN_TARGET])[0]
    args.extra_args = args._positionals[pos + has_sep :]

    show_plugin_help = "-h" in args.extra_args or "--help" in args.extra_args

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if not show_plugin_help and args.project_dir:
            raise

    if project is not None:
        env = project.env
    else:
        env = DEFAULT_ENVIRONMENT

    try:
        transform = env.transforms[args.transform]
    except KeyError:
        raise CliException("Transform '%s' is not found" % args.transform)

    extra_args = transform.parse_cmdline(args.extra_args)

    is_target = project is not None and args.target in project.working_tree.build_targets
    if is_target:
        if (
            not args.dst_dir
            and args.stage
            and (args.target != ProjectBuildTargets.strip_target_name(args.target))
        ):
            raise CliException(
                "Adding a stage is only allowed for " "project targets, not their stages."
            )

        if args.target == ProjectBuildTargets.MAIN_TARGET:
            targets = list(project.working_tree.sources)
        else:
            targets = [args.target]

        build_tree = project.working_tree.clone()
        for target in targets:
            build_tree.build_targets.add_transform_stage(target, args.transform, params=extra_args)

    if args.apply:
        log.info("Transforming...")

        if is_target and not args.dst_dir:
            for target in targets:
                dataset = project.working_tree.make_dataset(build_tree.make_pipeline(target))

                # Source might be missing in the working dir, so we specify
                # the output directory
                # We specify save_media here as a heuristic. It can probably
                # be improved by checking if there are images in the dataset
                # directory.
                dataset.save(project.source_data_dir(target), save_media=True)

            log.info("Finished")
        else:
            dataset, _project = parse_full_revpath(args.target, project)
            if _project:
                scope_add(_project)

            dst_dir = args.dst_dir or dataset.data_path
            if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
                raise CliException(
                    "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
                )
            dst_dir = osp.abspath(dst_dir)

            dataset.transform(args.transform, **extra_args)
            dataset.save(dst_dir, save_media=True)

            log.info("Results have been saved to '%s'" % dst_dir)

    if is_target and args.stage:
        project.working_tree.config.update(build_tree.config)
        project.working_tree.save()

    return 0



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# pylint: disable=no-self-use

import argparse
import logging as log
import os.path as osp
import struct
import warnings
from io import BufferedWriter
from multiprocessing.pool import ApplyResult, Pool
from typing import Any, List, Optional, Union

from datumaro.components.crypter import NULL_CRYPTER, Crypter
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.errors import DatumaroError
from datumaro.components.exporter import ExportContext, ExportContextComponent, Exporter
from datumaro.plugins.data_formats.datumaro.exporter import DatumaroExporter
from datumaro.plugins.data_formats.datumaro.exporter import _SubsetWriter as __SubsetWriter
from datumaro.plugins.data_formats.datumaro.format import DATUMARO_FORMAT_VERSION

from .format import DatumaroBinaryPath
from .mapper import DictMapper
from .mapper.common import IntListMapper
from .mapper.dataset_item import DatasetItemMapper


class _SubsetWriter(__SubsetWriter):
    """"""

    def __init__(
        self,
        context: Exporter,
        subset: str,
        ann_file: str,
        export_context: ExportContextComponent,
        secret_key_file: str,
        no_media_encryption: bool = False,
        max_blob_size: int = DatumaroBinaryPath.MAX_BLOB_SIZE,
    ):
        super().__init__(context, subset, ann_file, export_context)
        self._crypter = self.export_context.crypter
        self.secret_key_file = secret_key_file

        self._fp: Optional[BufferedWriter] = None
        self._data["items"]: List[Union[bytes, ApplyResult]] = []
        self._bytes: List[Union[bytes, ApplyResult]] = self._data["items"]
        self._item_cnt = 0
        media_type = context._extractor.media_type()
        self._media_type = {"media_type": media_type._type}

        self._media_encryption = not no_media_encryption

        if max_blob_size != DatumaroBinaryPath.MAX_BLOB_SIZE:
            warnings.warn(
                f"You provide max_blob_size={max_blob_size}, "
                "but it is not recommended to provide an arbitrary max_blob_size."
            )

        self._max_blob_size = max_blob_size

    def _sign(self):
        self._fp.write(DatumaroBinaryPath.SIGNATURE.encode())

    def _dump_encryption_field(self) -> int:
        if self._crypter.key is None:
            msg = b""
        else:
            msg = self._crypter.key
            msg = self._crypter.encrypt(msg)

        return self._fp.write(struct.pack(f"I{len(msg)}s", len(msg), msg))

    def _dump_header(self, header: Any, use_crypter: bool = True):
        msg = DictMapper.forward(header)

        if use_crypter and self._crypter.key is not None:
            msg = self._crypter.encrypt(msg)

        length = struct.pack("I", len(msg))
        return self._fp.write(length + msg)

    def _dump_version(self):
        self._dump_header(
            {
                "dm_format_version": DATUMARO_FORMAT_VERSION,
                "media_encryption": self._media_encryption,
            },
            use_crypter=False,
        )

    def _dump_info(self):
        self._dump_header(self.infos)

    def _dump_categories(self):
        self._dump_header(self.categories)

    def _dump_media_type(self):
        self._dump_header(self._media_type)

    def add_item(self, item: DatasetItem, pool: Optional[Pool] = None, *args, **kwargs):
        if pool is not None:
            self._bytes.append(
                pool.apply_async(
                    self.add_item_impl,
                    (
                        item,
                        self.export_context,
                        self._media_encryption,
                    ),
                )
            )
        else:
            self._bytes.append(
                self.add_item_impl(item, self.export_context, self._media_encryption)
            )

        self._item_cnt += 1

    @staticmethod
    def add_item_impl(
        item: DatasetItem, context: ExportContextComponent, media_encryption: bool
    ) -> bytes:
        with _SubsetWriter.context_save_media(item, context=context, encryption=media_encryption):
            return DatasetItemMapper.forward(item)

    def _dump_items(self, pool: Optional[Pool] = None):
        # Await async results
        if pool is not None:
            self._bytes = [
                result.get(timeout=DatumaroBinaryPath.MP_TIMEOUT)
                for result in self._bytes
                if isinstance(result, ApplyResult)
            ]

        # Divide items to blobs
        blobs = [bytearray()]
        cur_blob = blobs[-1]
        for _bytes in self._bytes:
            cur_blob += _bytes

            if len(cur_blob) > self._max_blob_size:
                blobs += [bytearray()]
                cur_blob = blobs[-1]

        # Encrypt blobs
        blobs = [self._crypter.encrypt(bytes(blob)) for blob in blobs if len(blob) > 0]

        # Dump blob sizes first
        blob_sizes = IntListMapper.forward([len(blob) for blob in blobs])
        blob_sizes = self._crypter.encrypt(blob_sizes)
        n_blob_sizes = len(blob_sizes)
        self._fp.write(struct.pack(f"<I{n_blob_sizes}s", n_blob_sizes, blob_sizes))

        # Dump blobs
        for blob in blobs:
            items_bytes = blob
            n_items_bytes = len(items_bytes)
            self._fp.write(struct.pack(f"<{n_items_bytes}s", items_bytes))

    def write(self, pool: Optional[Pool] = None, *args, **kwargs):
        try:
            if not self._crypter.is_null_crypter:
                log.info(
                    "Please see the generated encryption secret key file in the following path.\n"
                    f"{self.secret_key_file}\n"
                    "It must be kept it separate from the dataset to protect your dataset safely. "
                    "You also need it to import the encrpted dataset in later, so that be careful not to lose."
                )

                with open(self.secret_key_file, "w") as fp:
                    fp.write(self._crypter.key.decode())

            with open(self.ann_file, "wb") as fp:
                self._fp = fp
                self._sign()
                self._dump_version()
                self._dump_encryption_field()
                self._dump_info()
                self._dump_categories()
                self._dump_media_type()
                self._dump_items(pool)
        finally:
            self._fp = None


class EncryptionAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        encryption = True if option_string in self.option_strings else False
        if encryption:
            key = Crypter.gen_key()
        else:
            key = None

        setattr(namespace, "encryption_key", key)
        delattr(namespace, self.dest)


class DatumaroBinaryExporter(DatumaroExporter):
    DEFAULT_IMAGE_EXT = DatumaroBinaryPath.IMAGE_EXT
    PATH_CLS = DatumaroBinaryPath

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)

        parser.add_argument(
            "--encryption",
            action=EncryptionAction,
            default=False,
            help="Encrypt your dataset with the auto-generated secret key.",
        )

        parser.add_argument(
            "--no-media-encryption",
            action="store_true",
            help="Only encrypt the annotation file, not media files. "
            'This option is effective only if "--encryption" is enabled.',
        )

        parser.add_argument(
            "--num-workers",
            type=int,
            default=0,
            help="The number of multi-processing workers for export. "
            "If num_workers = 0, do not use multiprocessing (default: %(default)s).",
        )

        return parser

    def __init__(
        self,
        extractor: IDataset,
        save_dir: str,
        *,
        save_media: Optional[bool] = None,
        image_ext: Optional[str] = None,
        default_image_ext: Optional[str] = None,
        save_dataset_meta: bool = False,
        ctx: Optional[ExportContext] = None,
        encryption_key: Optional[bytes] = None,
        no_media_encryption: bool = False,
        encryption: bool = False,
        num_workers: int = 0,
        max_blob_size: int = DatumaroBinaryPath.MAX_BLOB_SIZE,
        **kwargs,
    ):
        """
        Parameters
        ----------
        encryption_key
            If provided, the dataset is encrypted with this key for export.
        no_media_encryption
            If true and encryption is enabled, do not encrypt media files and only encrypt annotation files.
        encryption
            If true and encryption_key is None, generate a random secret key.
        num_workers
            The number of multi-processing workers for export. If num_workers = 0, do not use multiprocessing.
        max_blob_size
            The maximum size of DatasetItem serialization blob. Changing from the default is not recommended.
        """

        if encryption and encryption_key is None:
            encryption_key = Crypter.gen_key()

        self._encryption_key = encryption_key

        if not save_media:
            no_media_encryption = True

        self._no_media_encryption = no_media_encryption

        if num_workers < 0:
            raise DatumaroError(
                f"num_workers should be non-negative but num_workers={num_workers}."
            )
        self._num_workers = num_workers

        self._max_blob_size = max_blob_size

        self._crypter = Crypter(encryption_key) if encryption_key is not None else NULL_CRYPTER

        super().__init__(
            extractor,
            save_dir,
            save_media=save_media,
            image_ext=image_ext,
            default_image_ext=default_image_ext,
            save_dataset_meta=save_dataset_meta,
            ctx=ctx,
        )

    def create_writer(
        self, subset: str, images_dir: str, pcd_dir: str, video_dir: str
    ) -> _SubsetWriter:
        export_context = ExportContextComponent(
            save_dir=self._save_dir,
            save_media=self._save_media,
            images_dir=images_dir,
            pcd_dir=pcd_dir,
            video_dir=video_dir,
            crypter=self._crypter,
            image_ext=self._image_ext,
            default_image_ext=self._default_image_ext,
        )

        return _SubsetWriter(
            context=self,
            subset=subset,
            ann_file=osp.join(self._annotations_dir, subset + self.PATH_CLS.ANNOTATION_EXT),
            export_context=export_context,
            secret_key_file=osp.join(self._save_dir, self.PATH_CLS.SECRET_KEY_FILE),
            no_media_encryption=self._no_media_encryption,
            max_blob_size=self._max_blob_size,
        )

    def _apply_impl(self, *args, **kwargs):
        if self._num_workers == 0:
            return super()._apply_impl()

        with Pool(processes=self._num_workers) as pool:
            return super()._apply_impl(pool)



# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT

import inspect
import logging as log
import os.path as osp
from collections import defaultdict
from importlib.util import module_from_spec, spec_from_file_location
from typing import Dict, List, Optional, Sequence, Tuple, overload

import numpy as np

from datumaro.components.abstracts.model_interpreter import (
    IModelInterpreter,
    LauncherInputType,
    ModelPred,
    PrepInfo,
)
from datumaro.components.annotation import Annotation
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetItem
from datumaro.errors import DatumaroError


# pylint: disable=no-self-use
class Launcher(CliPlugin):
    def __init__(self, model_dir: Optional[str] = None):
        pass

    def preprocess(self, item: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        """Preprocess single dataset item before launch()

        There are two output types:

        1. The output is `np.ndarray`. For example, it can be image data as `np.ndarray` with BGR format (H, W, C).
        In this step, you usually implement resizing, normalizing, or color channel conversion
        for your launcher (or model).

        2. The output is `Dict[str, np.ndarray]`. For example, it can be image and text pairs.
        Therefore, this can be used for the model having multi modality for image and text inputs.
        """
        raise NotImplementedError()

    @overload
    def infer(self, inputs: Dict[str, np.ndarray]) -> List[ModelPred]:
        """
        It executes the actual model inference for the inputs.

        Parameters:
            inputs: Dictionary of input data (its key is string and value is numpy array).
        Returns:
            List of model outputs
        """
        ...

    @overload
    def infer(self, inputs: np.ndarray) -> List[ModelPred]:
        """
        It executes the actual model inference for the inputs.

        Parameters:
            inputs: Batch of the numpy formatted input data
        Returns:
            List of model outputs
        """
        ...

    def infer(self, inputs: LauncherInputType) -> List[ModelPred]:
        raise NotImplementedError

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        raise NotImplementedError()

    def launch(self, batch: Sequence[DatasetItem], stack: bool = True) -> List[List[Annotation]]:
        """Launch to obtain the inference outputs of items.

        Parameters:
            inputs: batch of Datasetitems
            stack: If true, launch inference for the stacked input for the batch-wise dimension
                Otherwise, launch inference for each input.

        Returns:
            A list of annotation list. Each annotation list is mapped to the input
            :class:`DatasetItem`. These annotation list are pseudo-labels obtained by
            the model inference.
        """
        if len(batch) == 0:
            return []

        inputs = []
        inputs_info = []

        for item in batch:
            prep_inp, prep_info = self.preprocess(item)
            inputs.append(prep_inp)
            inputs_info.append(prep_info)

        if stack:
            if all(isinstance(inp, np.ndarray) for inp in inputs):
                preds = self.infer(np.stack(inputs, axis=0))
            elif all(isinstance(inp, dict) for inp in inputs):
                grouped_by_key = defaultdict(list)
                for inp in inputs:
                    for k, v in inp.items():
                        grouped_by_key[k].append(v)

                grouped_inputs = {k: np.stack(v, axis=0) for k, v in grouped_by_key.items()}
                preds = self.infer(grouped_inputs)
            else:
                actual = [type(inp) for inp in inputs]
                raise DatumaroError(f"Inputs should be np.ndarray or dict. but {actual}")
        else:
            preds = [self.infer(inp) for inp in inputs]

        return [self.postprocess(pred, info) for pred, info in zip(preds, inputs_info)]

    def infos(self):
        return None

    def categories(self):
        return None

    def type_check(self, item: DatasetItem) -> bool:
        """Check the media type of dataset item.

        If False, the item is excluded from the input batch.
        """
        return True


class LauncherWithModelInterpreter(Launcher):
    def __init__(self, model_interpreter_path: str):
        self._interpreter = self._load_interpreter(file_path=model_interpreter_path)

    def preprocess(self, item: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        """Preprocessing an input DatasetItem.

        Parameters:
            img: Input Datasetitem

        Returns:
            It returns a tuple of preprocessed input and preprocessing information.
            The preprocessing information will be used in the postprocessing step.
            One use case for this would be an affine transformation of the output bounding box
            obtained by object detection models. Input images for those models are usually resized
            to fit the model input dimensions. As a result, the postprocessing step should refine
            the model output bounding box to match the original input image size.
        """
        return self._interpreter.preprocess(item)

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        """Postprocess a model prediction.

        Parameters:
            pred: Model prediction
            info: Preprocessing information coming from the preprocessing step

        Returns:
            A list of annotations which is created from the model predictions
        """
        return self._interpreter.postprocess(pred, info)

    def _load_interpreter(self, file_path: str) -> IModelInterpreter:
        fname, _ = osp.splitext(osp.basename(file_path))
        spec = spec_from_file_location(fname, file_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        interps = []
        for name, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and issubclass(obj, IModelInterpreter)
                and obj is not IModelInterpreter
            ):
                log.info(f"Load {name} for model interpreter.")
                interps.append(obj)

        if len(interps) == 0:
            raise DatumaroError(f"{file_path} has no class derived from IModelInterpreter.")

        if len(interps) > 1:
            raise DatumaroError(
                f"{file_path} has more than two classes derived from IModelInterpreter ({interps}). "
                "There should be only one ModelInterpreter in the file."
            )

        return interps.pop()()

    def categories(self):
        return self._interpreter.get_categories()



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import json
import os
import os.path as osp
import re
import warnings
from typing import Dict, Optional, Type, TypeVar, Union

import numpy as np
import pandas as pd
from defusedxml import ElementTree

from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    ExtractedMask,
    Label,
    LabelCategories,
    MaskCategories,
)
from datumaro.components.dataset import DatasetItem
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME, DatasetBase, SubsetBase
from datumaro.components.errors import InvalidAnnotationError, InvalidFieldError, MissingFieldError
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image, ImageFromFile
from datumaro.plugins.data_formats.coco.base import CocoInstancesBase
from datumaro.plugins.data_formats.coco.format import CocoTask
from datumaro.plugins.data_formats.coco.page_mapper import COCOPageMapper
from datumaro.util import parse_json_file
from datumaro.util.image import IMAGE_EXTENSIONS, lazy_image

T = TypeVar("T")


class KaggleImageCsvBase(DatasetBase):
    def __init__(
        self,
        path: str,
        ann_file: str,
        columns: Dict[str, str],
        *,
        subset: Optional[str] = DEFAULT_SUBSET_NAME,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(ctx=ctx)

        self._subset = subset
        self._path = path

        if "media" not in columns:
            raise MissingFieldError("media")

        self._label_cat = LabelCategories()
        self._items = self._load_items(ann_file, columns)
        self._categories = {AnnotationType.label: self._label_cat}

    def _get_media_path(self, media_name: str):
        media_path = osp.join(self._path, media_name)
        if osp.exists(media_path):
            return media_path

        for ext in IMAGE_EXTENSIONS:
            media_path_with_ext = media_path + ext
            if osp.exists(media_path_with_ext):
                return media_path_with_ext

        return None

    def _parse_bbox_coords(self, bbox_str):
        coords = re.findall(r"[-+]?\d*\.\d+|\d+", bbox_str)
        if len(coords) != 4:
            raise ValueError("Bounding box coordinates must have exactly 4 values.")

        # expected to output [x1, y1, x2, y2]
        return [float(coord.strip()) for coord in coords]

    def _load_annotations(
        self, datas: list, indices: Dict[str, Union[int, Dict[str, int]]], bbox_flag: bool
    ):
        if "label" in indices:
            label_indices = indices["label"]
            if isinstance(label_indices, dict):
                labels = []
                list_values = datas[1:]
                index_to_label = {v: k for k, v in label_indices.items()}
                present_labels = [
                    index_to_label[i + 1] for i, value in enumerate(list_values) if value == "1"
                ]

                for label_name in present_labels:
                    label, cat = self._label_cat.find(label_name)
                    if not cat:
                        self._label_cat.add(label_name)
                        label, _ = self._label_cat.find(label_name)
                    labels.append(Label(label=label))
            else:
                label_name = str(datas[indices["label"]])
                label, cat = self._label_cat.find(label_name)
                if not cat:
                    self._label_cat.add(label_name)
                    label, _ = self._label_cat.find(label_name)
        else:
            _, cat = self._label_cat.find("object")
            if not cat:
                self._label_cat.add("object")
            label = 0

        if "label" in indices and not bbox_flag:
            label_indices = indices["label"]
            if isinstance(label_indices, dict):
                return labels
            return Label(label=label)

        if bbox_flag:
            if "bbox" in indices:
                coords = self._parse_bbox_coords(datas[indices["bbox"]])
                return Bbox(
                    label=label,
                    x=coords[0],
                    y=coords[1],
                    w=coords[2] - coords[0],
                    h=coords[3] - coords[1],
                )
            if "width" in indices and "height" in indices:
                return Bbox(
                    label=label,
                    x=float(datas[indices["x1"]]),
                    y=float(datas[indices["y1"]]),
                    w=float(datas[indices["width"]]),
                    h=float(datas[indices["height"]]),
                )
            if "x2" in indices and "y2" in indices:
                return Bbox(
                    label=label,
                    x=float(datas[indices["x1"]]),
                    y=float(datas[indices["y1"]]),
                    w=float(datas[indices["x2"]]) - float(datas[indices["x1"]]),
                    h=float(datas[indices["y2"]]) - float(datas[indices["y1"]]),
                )

    def _load_items(self, ann_file: str, columns: Dict[str, Union[str, list]]):
        df = pd.read_csv(ann_file, header=None, on_bad_lines="skip")
        df_fields = list(df.iloc[0])

        indices = {"media": df_fields.index(columns["media"])}
        if "label" in columns:
            label_columns = columns["label"]
            if isinstance(label_columns, list):
                indices_label = {}
                for label in label_columns:
                    indices_label[label] = df_fields.index(label)
                indices.update({"label": indices_label})
            else:
                indices.update({"label": df_fields.index(label_columns)})

        bbox_flag = False
        bbox_index = columns.get("bbox")
        if bbox_index:
            bbox_flag = True
            bbox_indices = {"x1", "x2", "y1", "y2", "width", "height"}
            if isinstance(bbox_index, str):
                indices["bbox"] = df_fields.index(bbox_index)
            elif isinstance(bbox_index, dict):
                indices.update(
                    {
                        key: df_fields.index(bbox_index[key])
                        for key in bbox_indices
                        if bbox_index.get(key)
                    }
                )
            if not (
                {"x1", "x2", "y1", "y2"} <= bbox_indices
                or {"x1", "y1", "width", "height"} <= bbox_indices
            ):
                warnings.warn("Insufficient box coordinate is given for importing bounding boxes.")
                bbox_flag = False

        items = dict()
        for _, row in df.iloc[1:].iterrows():  # Skip header row
            data_info = list(row)

            media_name = data_info[indices["media"]]
            item_id = osp.splitext(media_name)[0]

            media_path = self._get_media_path(media_name)
            if not media_path or not osp.exists(media_path):
                warnings.warn(
                    f"'{media_path}' is not existed in the directory, "
                    f"so we skip to create an dataset item according to {row}."
                )
                continue

            ann = self._load_annotations(data_info, indices, bbox_flag)
            if isinstance(ann, list):
                for label in ann:
                    self._ann_types.add(label.type)
                if item_id in items:
                    for label in ann:
                        items[item_id].annotations.append(label)
                else:
                    items[item_id] = DatasetItem(
                        id=item_id,
                        subset=self._subset,
                        media=Image.from_file(path=media_path),
                        annotations=ann,
                    )
            else:
                self._ann_types.add(ann.type)
                if item_id in items:
                    items[item_id].annotations.append(ann)
                else:
                    items[item_id] = DatasetItem(
                        id=item_id,
                        subset=self._subset,
                        media=Image.from_file(path=media_path),
                        annotations=[ann],
                    )
        return items.values()

    def categories(self):
        return self._categories

    def __iter__(self):
        yield from self._items

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--path", required=True)
        parser.add_argument("--ann_file", required=True)
        parser.add_argument("--columns", required=True, type=json.loads)

        return parser


class KaggleImageTxtBase(KaggleImageCsvBase):
    def __init__(
        self,
        path: str,
        ann_file: str,
        columns: Dict[str, int],
        *,
        subset: Optional[str] = DEFAULT_SUBSET_NAME,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(path=path, ann_file=ann_file, columns=columns, subset=subset, ctx=ctx)

    def _load_items(self, ann_file: str, columns: Dict[str, Union[int, Dict]]):
        bbox_flag = False
        if "bbox" in columns:
            bbox_flag = True
            bbox_columns = columns.pop("bbox")
            if isinstance(bbox_columns, dict):
                if not (
                    all(item in bbox_columns for item in ["x1", "x2", "y1", "y2"])
                    or all(item in bbox_columns for item in ["x1", "y1", "width", "height"])
                ):
                    warnings.warn(
                        "Insufficient box coordinate is given for importing bounding boxes."
                    )
                    bbox_flag = False
                columns.update(bbox_columns)

        items = dict()
        with open(ann_file, "r", encoding="utf-8") as f:
            for line in f:
                line = re.split(r"\s|,", line)

                media_name = line[columns["media"]]
                item_id = osp.splitext(media_name)[0]

                media_path = self._get_media_path(media_name)
                if not media_path or not osp.exists(media_path):
                    warnings.warn(
                        f"'{media_path}' is not existed in the directory, "
                        f"so we skip to create an dataset item according to {line}."
                    )
                    continue

                ann = self._load_annotations(line, columns, bbox_flag)
                self._ann_types.add(ann.type)
                if item_id in items:
                    items[item_id].annotations.append(ann)
                else:
                    items[item_id] = DatasetItem(
                        id=item_id,
                        subset=self._subset,
                        media=Image.from_file(path=media_path),
                        annotations=[ann],
                    )

        return items.values()


class KaggleImageMaskBase(DatasetBase):
    def __init__(
        self,
        path: str,
        mask_path: str,
        labelmap_file: Optional[str] = None,
        *,
        subset: Optional[str] = DEFAULT_SUBSET_NAME,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(ctx=ctx)

        self._subset = subset

        self._path = path
        self._mask_path = mask_path

        self._label_ids = []
        self._categories = self._load_categories(labelmap_file)
        self._items = self._load_items()

    def _load_categories(self, label_map_file: Optional[str]):
        label_map = dict()
        if not label_map_file:
            label_map["background"] = (0, 0, 0)
            label_map["object"] = (255, 255, 255)
        else:
            df = pd.read_csv(label_map_file)
            for _, row in df.iterrows():
                name = row[0]
                color = tuple([int(c) for c in row[1:]])
                label_map[name] = color

        label_categories = LabelCategories()
        for label in label_map:
            label_categories.add(label)

        categories = {}
        categories[AnnotationType.label] = label_categories

        colormap = {}
        for label_name, label_color in label_map.items():
            label_id = label_categories.find(label_name)[0]
            colormap[label_id] = label_color
            self._label_ids.append(label_id)

        categories[AnnotationType.mask] = MaskCategories(colormap)

        return categories

    def _load_items(self):
        items = []
        for media_name in sorted(os.listdir(self._path)):
            id = osp.splitext(media_name)[0]

            anns = []
            for mask_name in os.listdir(self._mask_path):
                if id in mask_name:
                    index_mask = lazy_image(
                        path=osp.join(self._mask_path, mask_name), dtype=np.int32
                    )
                    for label_id in self._label_ids:
                        anns.append(
                            ExtractedMask(
                                index_mask=index_mask,
                                index=label_id,
                                label=label_id,
                            )
                        )
                        self._ann_types.add(AnnotationType.mask)

            items.append(
                DatasetItem(
                    id=id,
                    subset=self._subset,
                    media=Image.from_file(path=osp.join(self._path, media_name)),
                    annotations=anns,
                )
            )

        return items

    def categories(self):
        return self._categories

    def __iter__(self):
        yield from self._items

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--path", required=True)
        parser.add_argument("--mask_path", required=True)
        parser.add_argument("--labelmap_file")

        return parser


class KaggleVocBase(SubsetBase):
    ann_extensions = ".xml"

    def __init__(
        self,
        path: str,
        ann_path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(subset=subset, ctx=ctx)

        self._label_cat = LabelCategories()
        self._items = []
        self._size = None

        for img_filename in sorted(os.listdir(path)):
            if not img_filename.lower().endswith(tuple(IMAGE_EXTENSIONS)):
                continue
            item_id = osp.splitext(img_filename)[0]

            img_file = osp.join(path, img_filename)
            ann_file = osp.join(ann_path, item_id + self.ann_extensions)

            annotations = (
                self._parse_annotations(img_file, ann_file) if osp.isfile(ann_file) else []
            )
            for ann in annotations:
                self._ann_types.add(ann.type)

            media = Image.from_file(path=img_file, size=self._size)

            self._items.append(
                DatasetItem(
                    id=item_id,
                    subset=self._subset,
                    media=media,
                    annotations=annotations,
                )
            )
        self._categories = {AnnotationType.label: self._label_cat}

    def _parse_annotations(self, img_file: str, ann_file: str):
        root_elem = ElementTree.parse(ann_file).getroot()
        if root_elem.tag != "annotation":
            raise MissingFieldError("annotation")

        height = self._parse_field(root_elem, "size/height", int, required=False)
        width = self._parse_field(root_elem, "size/width", int, required=False)
        if height and width:
            self._size = (height, width)

        annotations = []
        for obj_id, object_elem in enumerate(root_elem.iterfind("object")):
            label_name = self._parse_field(object_elem, "name", str, required=True)

            bbox_elem = object_elem.find("bndbox")
            if not bbox_elem:
                raise MissingFieldError("bndbox")

            xmin = self._parse_field(bbox_elem, "xmin", float)
            xmax = self._parse_field(bbox_elem, "xmax", float)
            ymin = self._parse_field(bbox_elem, "ymin", float)
            ymax = self._parse_field(bbox_elem, "ymax", float)

            label_id, cat = self._label_cat.find(label_name)
            if not cat:
                self._label_cat.add(label_name)
                label_id, _ = self._label_cat.find(label_name)

            annotations.append(
                Bbox(id=obj_id, label=label_id, x=xmin, y=ymin, w=xmax - xmin, h=ymax - ymin)
            )

        return annotations

    @staticmethod
    def _parse_field(root, xpath: str, cls: Type[T] = str, required: bool = True) -> Optional[T]:
        elem = root.find(xpath)
        if elem is None:
            if required:
                raise MissingFieldError(xpath)
            else:
                return None

        if cls is str:
            return elem.text

        try:
            return cls(elem.text)
        except Exception as e:
            raise InvalidFieldError(xpath) from e

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--path", required=True)
        parser.add_argument("--ann_path", required=True)
        parser.add_argument("--subset")

        return parser


class KaggleYoloBase(KaggleVocBase, SubsetBase):
    ann_extensions = ".txt"

    def __init__(
        self,
        path: str,
        ann_path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(path=path, ann_path=ann_path, subset=subset, ctx=ctx)

    def _parse_annotations(self, img_file: str, ann_file: str):
        image = ImageFromFile(path=img_file)
        image_height, image_width = image.size

        lines = []
        with open(ann_file, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if line:
                    lines.append(line)

        annotations = []
        for obj_id, line in enumerate(lines):
            parts = line.split()
            if len(parts) != 5:
                raise InvalidAnnotationError(
                    f"Unexpected field count {len(parts)} in the bbox description. "
                    "Expected 5 fields (label, xc, yc, w, h)."
                )
            label_name, xc, yc, w, h = parts
            xc = float(xc)
            yc = float(yc)
            w = float(w)
            h = float(h)
            x = (xc - w * 0.5) * image_width
            y = (yc - h * 0.5) * image_height
            w *= image_width
            h *= image_height

            label_id, cat = self._label_cat.find(label_name)
            if not cat:
                self._label_cat.add(label_name)
                label_id, _ = self._label_cat.find(label_name)
            label_id, _ = self._label_cat.find(label_name)

            annotations.append(Bbox(id=obj_id, label=label_id, x=x, y=y, w=w, h=h))

        return annotations


class KaggleCocoBase(CocoInstancesBase, SubsetBase):
    def __init__(
        self,
        path: str,
        ann_file: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
        stream: bool = False,
    ):
        SubsetBase.__init__(self, subset=subset, ctx=ctx)

        self._rootpath = path
        self._images_dir = path
        self._path = ann_file
        self._task = CocoTask.instances
        self._merge_instance_polygons = False

        keep_original_category_ids = False

        self._stream = stream
        if not stream:
            self._page_mapper = None  # No use in case of stream = False

            json_data = parse_json_file(ann_file)

            self._load_categories(
                json_data,
                keep_original_ids=keep_original_category_ids,
            )

            self._items = self._load_items(json_data)

            del json_data
        else:
            self._page_mapper = COCOPageMapper(ann_file)

            categories_data = self._page_mapper.stream_parse_categories_data()

            self._load_categories(
                {"categories": categories_data},
                keep_original_ids=keep_original_category_ids,
            )

            self._length = None

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument("--path", required=True)
        parser.add_argument("--ann_file", required=True)
        parser.add_argument("--subset")

        return parser



# Copyright (C) 2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod

from datumaro.components.cli_plugin import CliPlugin


class DatasetGenerator(ABC, CliPlugin):
    """
    ImageGenerator generates synthetic data with provided shape.
    """

    @classmethod
    def generate(cls, output_dir, count, shape, **options):
        generator = cls(output_dir, count, shape, **options)
        return generator.generate_dataset()

    @abstractmethod
    def generate_dataset(self):
        ...



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse

from datumaro.components.annotation import AnnotationType
from datumaro.components.errors import DatasetMergeError, MissingObjectError, ProjectNotFoundError
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.project import load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Prints dataset overview",
        description="""
        Prints info about the dataset at <revpath>, or about the current
        project's combined dataset, if none is specified.|n
        |n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        Both forms use the -p/--project as a context for plugins. It can be
        useful for dataset paths in targets. When not specified, the current
        project's working tree is used.|n
        |n
        Examples:|n
        - Print dataset info for the current project's working tree:|n
        |s|s%(prog)s|n
        |n
        - Print dataset info for a path and a format name:|n
        |s|s%(prog)s path/to/dataset:voc|n
        |n
        - Print dataset info for a source from a past revision:|n
        |s|s%(prog)s HEAD~2:source-2
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "target", nargs="?", default="project", metavar="revpath", help="Target dataset revpath"
    )
    parser.add_argument("--all", action="store_true", help="Print all information")
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the current project (default: current dir)",
    )
    parser.set_defaults(command=info_command)

    return parser


def get_sensitive_args():
    return {
        info_command: [
            "target",
            "project_dir",
        ],
    }


@scoped
def info_command(args):
    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    dataset = None
    dataset_problem = ""
    try:
        # TODO: avoid computing working tree hashes
        dataset, target_project = parse_full_revpath(args.target, project)
        if target_project:
            scope_add(target_project)
    except DatasetMergeError as e:
        dataset_problem = (
            "Can't merge project sources automatically: %s. "
            "The conflicting sources are: %s"
            % (
                e,
                ", ".join(e.sources),
            )
        )
    except MissingObjectError as e:
        dataset_problem = str(e)

    def print_dataset_info(dataset, indent=""):
        print("%slength:" % indent, len(dataset))

        categories = dataset.categories()
        print("%scategories:" % indent, ", ".join(c.name for c in categories))

        for cat_type, cat in categories.items():
            print("%s  %s:" % (indent, cat_type.name))
            if cat_type == AnnotationType.label:
                print("%s    count:" % indent, len(cat.items))

                count_threshold = 10
                if args.all:
                    count_threshold = len(cat.items)
                labels = ", ".join(c.name for c in cat.items[:count_threshold])
                if count_threshold < len(cat.items):
                    labels += " (and %s more)" % (len(cat.items) - count_threshold)
                print("%s    labels:" % indent, labels)

    if dataset is not None:
        print_dataset_info(dataset)

        subsets = dataset.subsets()
        print("subsets:", ", ".join(subsets))
        for subset_name in subsets:
            subset = dataset.get_subset(subset_name)
            print("  '%s':" % subset_name)
            print_dataset_info(subset, indent="    ")
    else:
        print("Dataset info is not available: ", dataset_problem)

    return 0



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional, Sequence, Type, TypeVar, Union, cast

import attr
from attr import attrs, field

from datumaro.components.annotation import Annotation, Annotations, AnnotationType, Categories
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.contexts.importer import ImportContext, NullImportContext
from datumaro.components.media import Image, MediaElement
from datumaro.util.attrs_util import default_if_none, not_empty
from datumaro.util.definitions import DEFAULT_SUBSET_NAME

T = TypeVar("T", bound=MediaElement)


@attrs(order=False, init=False, slots=True)
class DatasetItem:
    id: str = field(converter=lambda x: str(x).replace("\\", "/"), validator=not_empty)

    subset: str = field(converter=lambda v: v or DEFAULT_SUBSET_NAME, default=None)

    media: Optional[MediaElement] = field(
        default=None, validator=attr.validators.optional(attr.validators.instance_of(MediaElement))
    )

    annotations: Annotations = field(factory=Annotations, validator=default_if_none(Annotations))

    attributes: Dict[str, Any] = field(factory=dict, validator=default_if_none(dict))

    def wrap(item, **kwargs):
        return attr.evolve(item, **kwargs)

    def media_as(self, t: Type[T]) -> T:
        assert issubclass(t, MediaElement)
        return cast(t, self.media)

    def __init__(
        self,
        id: str,
        *,
        subset: Optional[str] = None,
        media: Union[str, MediaElement, None] = None,
        annotations: Optional[List[Annotation]] = None,
        attributes: Dict[str, Any] = None,
    ):
        self.__attrs_init__(
            id=id, subset=subset, media=media, annotations=annotations, attributes=attributes
        )


DatasetInfo = Dict[str, Any]
CategoriesInfo = Dict[AnnotationType, Categories]


class IDataset:
    def __iter__(self) -> Iterator[DatasetItem]:
        """
        Provides sequential access to dataset items.
        """
        raise NotImplementedError()

    def __len__(self) -> int:
        raise NotImplementedError()

    def __bool__(self):  # avoid __len__ use for truth checking
        return True

    def subsets(self) -> Dict[str, IDataset]:
        """
        Enumerates subsets in the dataset. Each subset can be a dataset itself.
        """
        raise NotImplementedError()

    def get_subset(self, name) -> IDataset:
        raise NotImplementedError()

    def infos(self) -> DatasetInfo:
        """
        Returns meta-info of dataset.
        """
        raise NotImplementedError()

    def categories(self) -> CategoriesInfo:
        """
        Returns metainfo about dataset labels.
        """
        raise NotImplementedError()

    def get(self, id: str, subset: Optional[str] = None) -> Optional[DatasetItem]:
        """
        Provides random access to dataset items.
        """
        raise NotImplementedError()

    def media_type(self) -> Type[MediaElement]:
        """
        Returns media type of the dataset items.

        All the items are supposed to have the same media type.
        Supposed to be constant and known immediately after the
        object construction (i.e. doesn't require dataset iteration).
        """
        raise NotImplementedError()

    def ann_types(self) -> List[AnnotationType]:
        """
        Returns available task type from dataset annotation types.
        """
        raise NotImplementedError()

    @property
    def is_stream(self) -> bool:
        """Boolean indicating whether the dataset is a stream

        If the dataset is a stream, the dataset item is generated on demand from its iterator.
        """
        return False


class _DatasetBase(IDataset):
    def __init__(self, *, length: Optional[int] = None, subsets: Optional[Sequence[str]] = None):
        self._length = length
        self._subsets = subsets
        self._ann_types = set()

    def _init_cache(self):
        subsets = set()
        length = -1
        for length, item in enumerate(self):
            subsets.add(item.subset)
        length += 1

        if self._length is None:
            self._length = length
        if self._subsets is None:
            self._subsets = subsets

    def __len__(self):
        if self._length is None:
            self._init_cache()
        return self._length

    def subsets(self) -> Dict[str, IDataset]:
        if self._subsets is None:
            self._init_cache()
        return {name or DEFAULT_SUBSET_NAME: self.get_subset(name) for name in self._subsets}

    def get_subset(self, name: str) -> IDataset:
        if self._subsets is None:
            self._init_cache()
        if name in self._subsets:
            if len(self._subsets) == 1:
                return self

            subset = self.select(lambda item: item.subset == name)
            subset._subsets = [name]
            return subset
        else:
            raise KeyError(
                "Unknown subset '%s', available subsets: %s" % (name, set(self._subsets))
            )

    def transform(self, method, *args, **kwargs):
        return method(self, *args, **kwargs)

    def select(self, pred):
        class _DatasetFilter(_DatasetBase):
            def __iter__(_):
                return filter(pred, iter(self))

            def infos(_):
                return self.infos()

            def categories(_):
                return self.categories()

            def media_type(_):
                return self.media_type()

            def ann_types(_):
                return self.ann_types()

        return _DatasetFilter()

    def infos(self) -> DatasetInfo:
        return {}

    def categories(self) -> CategoriesInfo:
        return {}

    def get(self, id, subset=None) -> Optional[DatasetItem]:
        subset = subset or DEFAULT_SUBSET_NAME
        for item in self:
            if item.id == id and item.subset == subset:
                return item
        return None


class DatasetBase(_DatasetBase, CliPlugin):
    """
    A base class for user-defined and built-in extractors.
    Should be used in cases, where SubsetBase is not enough,
    or its use makes problems with performance, implementation etc.
    """

    def __init__(
        self,
        *,
        length: Optional[int] = None,
        subsets: Optional[Sequence[str]] = None,
        media_type: Type[MediaElement] = Image,
        ann_types: Optional[List[AnnotationType]] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(length=length, subsets=subsets)

        self._ctx: ImportContext = ctx or NullImportContext()
        self._media_type = media_type
        self._ann_types = ann_types if ann_types else set()

    def media_type(self):
        return self._media_type

    def ann_types(self):
        return self._ann_types


class SubsetBase(DatasetBase):
    """
    A base class for simple, single-subset extractors.
    Should be used by default for user-defined extractors.
    """

    def __init__(
        self,
        *,
        length: Optional[int] = None,
        subset: Optional[str] = None,
        media_type: Type[MediaElement] = Image,
        ann_types: List[AnnotationType] = None,
        ctx: Optional[ImportContext] = None,
    ):
        self._subset = subset or DEFAULT_SUBSET_NAME
        super().__init__(
            length=length,
            subsets=[self._subset],
            media_type=media_type,
            ann_types=ann_types,
            ctx=ctx,
        )

        self._infos = {}
        self._categories = {}
        self._items = []

    def infos(self):
        return self._infos

    def categories(self):
        return self._categories

    def __iter__(self):
        yield from self._items

    def __len__(self):
        return len(self._items)

    def get(self, id, subset=None):
        assert subset == self._subset, "%s != %s" % (subset, self._subset)
        return super().get(id, subset or self._subset)

    @property
    def subset(self) -> str:
        """Subset name of this instance."""
        return self._subset
